# Generated from fugue_sql/_antlr/fugue_sql.g4 by ANTLR 4.9
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .fugue_sqlParser import fugue_sqlParser
else:
    from fugue_sqlParser import fugue_sqlParser

# This class defines a complete generic visitor for a parse tree produced by fugue_sqlParser.

class fugue_sqlVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by fugue_sqlParser#fugueLanguage.
    def visitFugueLanguage(self, ctx:fugue_sqlParser.FugueLanguageContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueSingleStatement.
    def visitFugueSingleStatement(self, ctx:fugue_sqlParser.FugueSingleStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueSingleTask.
    def visitFugueSingleTask(self, ctx:fugue_sqlParser.FugueSingleTaskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueNestableTask.
    def visitFugueNestableTask(self, ctx:fugue_sqlParser.FugueNestableTaskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueNestableTaskCollectionNoSelect.
    def visitFugueNestableTaskCollectionNoSelect(self, ctx:fugue_sqlParser.FugueNestableTaskCollectionNoSelectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueEngineSpecificQueryTask.
    def visitFugueEngineSpecificQueryTask(self, ctx:fugue_sqlParser.FugueEngineSpecificQueryTaskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueTransformTask.
    def visitFugueTransformTask(self, ctx:fugue_sqlParser.FugueTransformTaskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueProcessTask.
    def visitFugueProcessTask(self, ctx:fugue_sqlParser.FugueProcessTaskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueSaveAndUseTask.
    def visitFugueSaveAndUseTask(self, ctx:fugue_sqlParser.FugueSaveAndUseTaskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueRenameColumnsTask.
    def visitFugueRenameColumnsTask(self, ctx:fugue_sqlParser.FugueRenameColumnsTaskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueAlterColumnsTask.
    def visitFugueAlterColumnsTask(self, ctx:fugue_sqlParser.FugueAlterColumnsTaskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueDropColumnsTask.
    def visitFugueDropColumnsTask(self, ctx:fugue_sqlParser.FugueDropColumnsTaskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueDropnaTask.
    def visitFugueDropnaTask(self, ctx:fugue_sqlParser.FugueDropnaTaskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueFillnaTask.
    def visitFugueFillnaTask(self, ctx:fugue_sqlParser.FugueFillnaTaskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueSampleTask.
    def visitFugueSampleTask(self, ctx:fugue_sqlParser.FugueSampleTaskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueZipTask.
    def visitFugueZipTask(self, ctx:fugue_sqlParser.FugueZipTaskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueCreateTask.
    def visitFugueCreateTask(self, ctx:fugue_sqlParser.FugueCreateTaskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueCreateDataTask.
    def visitFugueCreateDataTask(self, ctx:fugue_sqlParser.FugueCreateDataTaskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueLoadTask.
    def visitFugueLoadTask(self, ctx:fugue_sqlParser.FugueLoadTaskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueOutputTask.
    def visitFugueOutputTask(self, ctx:fugue_sqlParser.FugueOutputTaskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fuguePrintTask.
    def visitFuguePrintTask(self, ctx:fugue_sqlParser.FuguePrintTaskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueSaveTask.
    def visitFugueSaveTask(self, ctx:fugue_sqlParser.FugueSaveTaskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueOutputTransformTask.
    def visitFugueOutputTransformTask(self, ctx:fugue_sqlParser.FugueOutputTransformTaskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueSqlEngine.
    def visitFugueSqlEngine(self, ctx:fugue_sqlParser.FugueSqlEngineContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueSingleFile.
    def visitFugueSingleFile(self, ctx:fugue_sqlParser.FugueSingleFileContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueLoadColumns.
    def visitFugueLoadColumns(self, ctx:fugue_sqlParser.FugueLoadColumnsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueSaveMode.
    def visitFugueSaveMode(self, ctx:fugue_sqlParser.FugueSaveModeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueFileFormat.
    def visitFugueFileFormat(self, ctx:fugue_sqlParser.FugueFileFormatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fuguePath.
    def visitFuguePath(self, ctx:fugue_sqlParser.FuguePathContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueCheckpointWeak.
    def visitFugueCheckpointWeak(self, ctx:fugue_sqlParser.FugueCheckpointWeakContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueCheckpointStrong.
    def visitFugueCheckpointStrong(self, ctx:fugue_sqlParser.FugueCheckpointStrongContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueCheckpointDeterministic.
    def visitFugueCheckpointDeterministic(self, ctx:fugue_sqlParser.FugueCheckpointDeterministicContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueCheckpointYield.
    def visitFugueCheckpointYield(self, ctx:fugue_sqlParser.FugueCheckpointYieldContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueCheckpointNamespace.
    def visitFugueCheckpointNamespace(self, ctx:fugue_sqlParser.FugueCheckpointNamespaceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueYield.
    def visitFugueYield(self, ctx:fugue_sqlParser.FugueYieldContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueBroadcast.
    def visitFugueBroadcast(self, ctx:fugue_sqlParser.FugueBroadcastContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueDataFramesList.
    def visitFugueDataFramesList(self, ctx:fugue_sqlParser.FugueDataFramesListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueDataFramesDict.
    def visitFugueDataFramesDict(self, ctx:fugue_sqlParser.FugueDataFramesDictContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueDataFramePair.
    def visitFugueDataFramePair(self, ctx:fugue_sqlParser.FugueDataFramePairContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueDataFrameSource.
    def visitFugueDataFrameSource(self, ctx:fugue_sqlParser.FugueDataFrameSourceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueDataFrameNested.
    def visitFugueDataFrameNested(self, ctx:fugue_sqlParser.FugueDataFrameNestedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueAssignment.
    def visitFugueAssignment(self, ctx:fugue_sqlParser.FugueAssignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueAssignmentSign.
    def visitFugueAssignmentSign(self, ctx:fugue_sqlParser.FugueAssignmentSignContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueSingleOutputExtensionCommonWild.
    def visitFugueSingleOutputExtensionCommonWild(self, ctx:fugue_sqlParser.FugueSingleOutputExtensionCommonWildContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueSingleOutputExtensionCommon.
    def visitFugueSingleOutputExtensionCommon(self, ctx:fugue_sqlParser.FugueSingleOutputExtensionCommonContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueExtension.
    def visitFugueExtension(self, ctx:fugue_sqlParser.FugueExtensionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueSampleMethod.
    def visitFugueSampleMethod(self, ctx:fugue_sqlParser.FugueSampleMethodContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueZipType.
    def visitFugueZipType(self, ctx:fugue_sqlParser.FugueZipTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fuguePrepartition.
    def visitFuguePrepartition(self, ctx:fugue_sqlParser.FuguePrepartitionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fuguePartitionAlgo.
    def visitFuguePartitionAlgo(self, ctx:fugue_sqlParser.FuguePartitionAlgoContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fuguePartitionNum.
    def visitFuguePartitionNum(self, ctx:fugue_sqlParser.FuguePartitionNumContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fuguePartitionNumber.
    def visitFuguePartitionNumber(self, ctx:fugue_sqlParser.FuguePartitionNumberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueParamsPairs.
    def visitFugueParamsPairs(self, ctx:fugue_sqlParser.FugueParamsPairsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueParamsObj.
    def visitFugueParamsObj(self, ctx:fugue_sqlParser.FugueParamsObjContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueCols.
    def visitFugueCols(self, ctx:fugue_sqlParser.FugueColsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueColsSort.
    def visitFugueColsSort(self, ctx:fugue_sqlParser.FugueColsSortContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueColSort.
    def visitFugueColSort(self, ctx:fugue_sqlParser.FugueColSortContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueColumnIdentifier.
    def visitFugueColumnIdentifier(self, ctx:fugue_sqlParser.FugueColumnIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueRenameExpression.
    def visitFugueRenameExpression(self, ctx:fugue_sqlParser.FugueRenameExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueWildSchema.
    def visitFugueWildSchema(self, ctx:fugue_sqlParser.FugueWildSchemaContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueWildSchemaPair.
    def visitFugueWildSchemaPair(self, ctx:fugue_sqlParser.FugueWildSchemaPairContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueSchema.
    def visitFugueSchema(self, ctx:fugue_sqlParser.FugueSchemaContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueSchemaPair.
    def visitFugueSchemaPair(self, ctx:fugue_sqlParser.FugueSchemaPairContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueSchemaKey.
    def visitFugueSchemaKey(self, ctx:fugue_sqlParser.FugueSchemaKeyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueSchemaSimpleType.
    def visitFugueSchemaSimpleType(self, ctx:fugue_sqlParser.FugueSchemaSimpleTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueSchemaListType.
    def visitFugueSchemaListType(self, ctx:fugue_sqlParser.FugueSchemaListTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueSchemaStructType.
    def visitFugueSchemaStructType(self, ctx:fugue_sqlParser.FugueSchemaStructTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueRenamePair.
    def visitFugueRenamePair(self, ctx:fugue_sqlParser.FugueRenamePairContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueJson.
    def visitFugueJson(self, ctx:fugue_sqlParser.FugueJsonContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueJsonObj.
    def visitFugueJsonObj(self, ctx:fugue_sqlParser.FugueJsonObjContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueJsonPairs.
    def visitFugueJsonPairs(self, ctx:fugue_sqlParser.FugueJsonPairsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueJsonPair.
    def visitFugueJsonPair(self, ctx:fugue_sqlParser.FugueJsonPairContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueJsonKey.
    def visitFugueJsonKey(self, ctx:fugue_sqlParser.FugueJsonKeyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueJsonArray.
    def visitFugueJsonArray(self, ctx:fugue_sqlParser.FugueJsonArrayContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueJsonValue.
    def visitFugueJsonValue(self, ctx:fugue_sqlParser.FugueJsonValueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueJsonNumber.
    def visitFugueJsonNumber(self, ctx:fugue_sqlParser.FugueJsonNumberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueJsonString.
    def visitFugueJsonString(self, ctx:fugue_sqlParser.FugueJsonStringContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueJsonBool.
    def visitFugueJsonBool(self, ctx:fugue_sqlParser.FugueJsonBoolContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueJsonNull.
    def visitFugueJsonNull(self, ctx:fugue_sqlParser.FugueJsonNullContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueIdentifier.
    def visitFugueIdentifier(self, ctx:fugue_sqlParser.FugueIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#singleStatement.
    def visitSingleStatement(self, ctx:fugue_sqlParser.SingleStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#singleExpression.
    def visitSingleExpression(self, ctx:fugue_sqlParser.SingleExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#singleTableIdentifier.
    def visitSingleTableIdentifier(self, ctx:fugue_sqlParser.SingleTableIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#singleMultipartIdentifier.
    def visitSingleMultipartIdentifier(self, ctx:fugue_sqlParser.SingleMultipartIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#singleFunctionIdentifier.
    def visitSingleFunctionIdentifier(self, ctx:fugue_sqlParser.SingleFunctionIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#singleDataType.
    def visitSingleDataType(self, ctx:fugue_sqlParser.SingleDataTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#singleTableSchema.
    def visitSingleTableSchema(self, ctx:fugue_sqlParser.SingleTableSchemaContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#statementDefault.
    def visitStatementDefault(self, ctx:fugue_sqlParser.StatementDefaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#dmlStatement.
    def visitDmlStatement(self, ctx:fugue_sqlParser.DmlStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#use.
    def visitUse(self, ctx:fugue_sqlParser.UseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#createNamespace.
    def visitCreateNamespace(self, ctx:fugue_sqlParser.CreateNamespaceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#setNamespaceProperties.
    def visitSetNamespaceProperties(self, ctx:fugue_sqlParser.SetNamespacePropertiesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#setNamespaceLocation.
    def visitSetNamespaceLocation(self, ctx:fugue_sqlParser.SetNamespaceLocationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#dropNamespace.
    def visitDropNamespace(self, ctx:fugue_sqlParser.DropNamespaceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#showNamespaces.
    def visitShowNamespaces(self, ctx:fugue_sqlParser.ShowNamespacesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#createTable.
    def visitCreateTable(self, ctx:fugue_sqlParser.CreateTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#createHiveTable.
    def visitCreateHiveTable(self, ctx:fugue_sqlParser.CreateHiveTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#createTableLike.
    def visitCreateTableLike(self, ctx:fugue_sqlParser.CreateTableLikeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#replaceTable.
    def visitReplaceTable(self, ctx:fugue_sqlParser.ReplaceTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#analyze.
    def visitAnalyze(self, ctx:fugue_sqlParser.AnalyzeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#addTableColumns.
    def visitAddTableColumns(self, ctx:fugue_sqlParser.AddTableColumnsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#renameTableColumn.
    def visitRenameTableColumn(self, ctx:fugue_sqlParser.RenameTableColumnContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#dropTableColumns.
    def visitDropTableColumns(self, ctx:fugue_sqlParser.DropTableColumnsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#renameTable.
    def visitRenameTable(self, ctx:fugue_sqlParser.RenameTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#setTableProperties.
    def visitSetTableProperties(self, ctx:fugue_sqlParser.SetTablePropertiesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#unsetTableProperties.
    def visitUnsetTableProperties(self, ctx:fugue_sqlParser.UnsetTablePropertiesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#alterTableAlterColumn.
    def visitAlterTableAlterColumn(self, ctx:fugue_sqlParser.AlterTableAlterColumnContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#hiveChangeColumn.
    def visitHiveChangeColumn(self, ctx:fugue_sqlParser.HiveChangeColumnContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#hiveReplaceColumns.
    def visitHiveReplaceColumns(self, ctx:fugue_sqlParser.HiveReplaceColumnsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#setTableSerDe.
    def visitSetTableSerDe(self, ctx:fugue_sqlParser.SetTableSerDeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#addTablePartition.
    def visitAddTablePartition(self, ctx:fugue_sqlParser.AddTablePartitionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#renameTablePartition.
    def visitRenameTablePartition(self, ctx:fugue_sqlParser.RenameTablePartitionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#dropTablePartitions.
    def visitDropTablePartitions(self, ctx:fugue_sqlParser.DropTablePartitionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#setTableLocation.
    def visitSetTableLocation(self, ctx:fugue_sqlParser.SetTableLocationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#recoverPartitions.
    def visitRecoverPartitions(self, ctx:fugue_sqlParser.RecoverPartitionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#dropTable.
    def visitDropTable(self, ctx:fugue_sqlParser.DropTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#dropView.
    def visitDropView(self, ctx:fugue_sqlParser.DropViewContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#createView.
    def visitCreateView(self, ctx:fugue_sqlParser.CreateViewContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#createTempViewUsing.
    def visitCreateTempViewUsing(self, ctx:fugue_sqlParser.CreateTempViewUsingContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#alterViewQuery.
    def visitAlterViewQuery(self, ctx:fugue_sqlParser.AlterViewQueryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#createFunction.
    def visitCreateFunction(self, ctx:fugue_sqlParser.CreateFunctionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#dropFunction.
    def visitDropFunction(self, ctx:fugue_sqlParser.DropFunctionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#explain.
    def visitExplain(self, ctx:fugue_sqlParser.ExplainContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#showTables.
    def visitShowTables(self, ctx:fugue_sqlParser.ShowTablesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#showTable.
    def visitShowTable(self, ctx:fugue_sqlParser.ShowTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#showTblProperties.
    def visitShowTblProperties(self, ctx:fugue_sqlParser.ShowTblPropertiesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#showColumns.
    def visitShowColumns(self, ctx:fugue_sqlParser.ShowColumnsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#showViews.
    def visitShowViews(self, ctx:fugue_sqlParser.ShowViewsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#showPartitions.
    def visitShowPartitions(self, ctx:fugue_sqlParser.ShowPartitionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#showFunctions.
    def visitShowFunctions(self, ctx:fugue_sqlParser.ShowFunctionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#showCreateTable.
    def visitShowCreateTable(self, ctx:fugue_sqlParser.ShowCreateTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#showCurrentNamespace.
    def visitShowCurrentNamespace(self, ctx:fugue_sqlParser.ShowCurrentNamespaceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#describeFunction.
    def visitDescribeFunction(self, ctx:fugue_sqlParser.DescribeFunctionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#describeNamespace.
    def visitDescribeNamespace(self, ctx:fugue_sqlParser.DescribeNamespaceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#describeRelation.
    def visitDescribeRelation(self, ctx:fugue_sqlParser.DescribeRelationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#describeQuery.
    def visitDescribeQuery(self, ctx:fugue_sqlParser.DescribeQueryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#commentNamespace.
    def visitCommentNamespace(self, ctx:fugue_sqlParser.CommentNamespaceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#commentTable.
    def visitCommentTable(self, ctx:fugue_sqlParser.CommentTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#refreshTable.
    def visitRefreshTable(self, ctx:fugue_sqlParser.RefreshTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#refreshResource.
    def visitRefreshResource(self, ctx:fugue_sqlParser.RefreshResourceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#cacheTable.
    def visitCacheTable(self, ctx:fugue_sqlParser.CacheTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#uncacheTable.
    def visitUncacheTable(self, ctx:fugue_sqlParser.UncacheTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#clearCache.
    def visitClearCache(self, ctx:fugue_sqlParser.ClearCacheContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#loadData.
    def visitLoadData(self, ctx:fugue_sqlParser.LoadDataContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#truncateTable.
    def visitTruncateTable(self, ctx:fugue_sqlParser.TruncateTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#repairTable.
    def visitRepairTable(self, ctx:fugue_sqlParser.RepairTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#manageResource.
    def visitManageResource(self, ctx:fugue_sqlParser.ManageResourceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#failNativeCommand.
    def visitFailNativeCommand(self, ctx:fugue_sqlParser.FailNativeCommandContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#setConfiguration.
    def visitSetConfiguration(self, ctx:fugue_sqlParser.SetConfigurationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#resetConfiguration.
    def visitResetConfiguration(self, ctx:fugue_sqlParser.ResetConfigurationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#unsupportedHiveNativeCommands.
    def visitUnsupportedHiveNativeCommands(self, ctx:fugue_sqlParser.UnsupportedHiveNativeCommandsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#createTableHeader.
    def visitCreateTableHeader(self, ctx:fugue_sqlParser.CreateTableHeaderContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#replaceTableHeader.
    def visitReplaceTableHeader(self, ctx:fugue_sqlParser.ReplaceTableHeaderContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#bucketSpec.
    def visitBucketSpec(self, ctx:fugue_sqlParser.BucketSpecContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#skewSpec.
    def visitSkewSpec(self, ctx:fugue_sqlParser.SkewSpecContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#locationSpec.
    def visitLocationSpec(self, ctx:fugue_sqlParser.LocationSpecContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#commentSpec.
    def visitCommentSpec(self, ctx:fugue_sqlParser.CommentSpecContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#query.
    def visitQuery(self, ctx:fugue_sqlParser.QueryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#insertOverwriteTable.
    def visitInsertOverwriteTable(self, ctx:fugue_sqlParser.InsertOverwriteTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#insertIntoTable.
    def visitInsertIntoTable(self, ctx:fugue_sqlParser.InsertIntoTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#insertOverwriteHiveDir.
    def visitInsertOverwriteHiveDir(self, ctx:fugue_sqlParser.InsertOverwriteHiveDirContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#insertOverwriteDir.
    def visitInsertOverwriteDir(self, ctx:fugue_sqlParser.InsertOverwriteDirContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#partitionSpecLocation.
    def visitPartitionSpecLocation(self, ctx:fugue_sqlParser.PartitionSpecLocationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#partitionSpec.
    def visitPartitionSpec(self, ctx:fugue_sqlParser.PartitionSpecContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#partitionVal.
    def visitPartitionVal(self, ctx:fugue_sqlParser.PartitionValContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#namespace.
    def visitNamespace(self, ctx:fugue_sqlParser.NamespaceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#describeFuncName.
    def visitDescribeFuncName(self, ctx:fugue_sqlParser.DescribeFuncNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#describeColName.
    def visitDescribeColName(self, ctx:fugue_sqlParser.DescribeColNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#ctes.
    def visitCtes(self, ctx:fugue_sqlParser.CtesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#namedQuery.
    def visitNamedQuery(self, ctx:fugue_sqlParser.NamedQueryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#tableProvider.
    def visitTableProvider(self, ctx:fugue_sqlParser.TableProviderContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#createTableClauses.
    def visitCreateTableClauses(self, ctx:fugue_sqlParser.CreateTableClausesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#tablePropertyList.
    def visitTablePropertyList(self, ctx:fugue_sqlParser.TablePropertyListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#tableProperty.
    def visitTableProperty(self, ctx:fugue_sqlParser.TablePropertyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#tablePropertyKey.
    def visitTablePropertyKey(self, ctx:fugue_sqlParser.TablePropertyKeyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#tablePropertyValue.
    def visitTablePropertyValue(self, ctx:fugue_sqlParser.TablePropertyValueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#constantList.
    def visitConstantList(self, ctx:fugue_sqlParser.ConstantListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#nestedConstantList.
    def visitNestedConstantList(self, ctx:fugue_sqlParser.NestedConstantListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#createFileFormat.
    def visitCreateFileFormat(self, ctx:fugue_sqlParser.CreateFileFormatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#tableFileFormat.
    def visitTableFileFormat(self, ctx:fugue_sqlParser.TableFileFormatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#genericFileFormat.
    def visitGenericFileFormat(self, ctx:fugue_sqlParser.GenericFileFormatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#storageHandler.
    def visitStorageHandler(self, ctx:fugue_sqlParser.StorageHandlerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#resource.
    def visitResource(self, ctx:fugue_sqlParser.ResourceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#singleInsertQuery.
    def visitSingleInsertQuery(self, ctx:fugue_sqlParser.SingleInsertQueryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#multiInsertQuery.
    def visitMultiInsertQuery(self, ctx:fugue_sqlParser.MultiInsertQueryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#deleteFromTable.
    def visitDeleteFromTable(self, ctx:fugue_sqlParser.DeleteFromTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#updateTable.
    def visitUpdateTable(self, ctx:fugue_sqlParser.UpdateTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#mergeIntoTable.
    def visitMergeIntoTable(self, ctx:fugue_sqlParser.MergeIntoTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#queryOrganization.
    def visitQueryOrganization(self, ctx:fugue_sqlParser.QueryOrganizationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#multiInsertQueryBody.
    def visitMultiInsertQueryBody(self, ctx:fugue_sqlParser.MultiInsertQueryBodyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#queryTermDefault.
    def visitQueryTermDefault(self, ctx:fugue_sqlParser.QueryTermDefaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fugueTerm.
    def visitFugueTerm(self, ctx:fugue_sqlParser.FugueTermContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#setOperation.
    def visitSetOperation(self, ctx:fugue_sqlParser.SetOperationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#queryPrimaryDefault.
    def visitQueryPrimaryDefault(self, ctx:fugue_sqlParser.QueryPrimaryDefaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fromStmt.
    def visitFromStmt(self, ctx:fugue_sqlParser.FromStmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#table.
    def visitTable(self, ctx:fugue_sqlParser.TableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#inlineTableDefault1.
    def visitInlineTableDefault1(self, ctx:fugue_sqlParser.InlineTableDefault1Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#sortItem.
    def visitSortItem(self, ctx:fugue_sqlParser.SortItemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fromStatement.
    def visitFromStatement(self, ctx:fugue_sqlParser.FromStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fromStatementBody.
    def visitFromStatementBody(self, ctx:fugue_sqlParser.FromStatementBodyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#transformQuerySpecification.
    def visitTransformQuerySpecification(self, ctx:fugue_sqlParser.TransformQuerySpecificationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#regularQuerySpecification.
    def visitRegularQuerySpecification(self, ctx:fugue_sqlParser.RegularQuerySpecificationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#optionalFromClause.
    def visitOptionalFromClause(self, ctx:fugue_sqlParser.OptionalFromClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#transformClause.
    def visitTransformClause(self, ctx:fugue_sqlParser.TransformClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#selectClause.
    def visitSelectClause(self, ctx:fugue_sqlParser.SelectClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#setClause.
    def visitSetClause(self, ctx:fugue_sqlParser.SetClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#matchedClause.
    def visitMatchedClause(self, ctx:fugue_sqlParser.MatchedClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#notMatchedClause.
    def visitNotMatchedClause(self, ctx:fugue_sqlParser.NotMatchedClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#matchedAction.
    def visitMatchedAction(self, ctx:fugue_sqlParser.MatchedActionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#notMatchedAction.
    def visitNotMatchedAction(self, ctx:fugue_sqlParser.NotMatchedActionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#assignmentList.
    def visitAssignmentList(self, ctx:fugue_sqlParser.AssignmentListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#assignment.
    def visitAssignment(self, ctx:fugue_sqlParser.AssignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#whereClause.
    def visitWhereClause(self, ctx:fugue_sqlParser.WhereClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#havingClause.
    def visitHavingClause(self, ctx:fugue_sqlParser.HavingClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#hint.
    def visitHint(self, ctx:fugue_sqlParser.HintContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#hintStatement.
    def visitHintStatement(self, ctx:fugue_sqlParser.HintStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#fromClause.
    def visitFromClause(self, ctx:fugue_sqlParser.FromClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#aggregationClause.
    def visitAggregationClause(self, ctx:fugue_sqlParser.AggregationClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#groupingSet.
    def visitGroupingSet(self, ctx:fugue_sqlParser.GroupingSetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#pivotClause.
    def visitPivotClause(self, ctx:fugue_sqlParser.PivotClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#pivotColumn.
    def visitPivotColumn(self, ctx:fugue_sqlParser.PivotColumnContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#pivotValue.
    def visitPivotValue(self, ctx:fugue_sqlParser.PivotValueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#lateralView.
    def visitLateralView(self, ctx:fugue_sqlParser.LateralViewContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#setQuantifier.
    def visitSetQuantifier(self, ctx:fugue_sqlParser.SetQuantifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#relation.
    def visitRelation(self, ctx:fugue_sqlParser.RelationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#joinRelation.
    def visitJoinRelation(self, ctx:fugue_sqlParser.JoinRelationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#joinType.
    def visitJoinType(self, ctx:fugue_sqlParser.JoinTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#joinCriteria.
    def visitJoinCriteria(self, ctx:fugue_sqlParser.JoinCriteriaContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#sample.
    def visitSample(self, ctx:fugue_sqlParser.SampleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#sampleByPercentile.
    def visitSampleByPercentile(self, ctx:fugue_sqlParser.SampleByPercentileContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#sampleByRows.
    def visitSampleByRows(self, ctx:fugue_sqlParser.SampleByRowsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#sampleByBucket.
    def visitSampleByBucket(self, ctx:fugue_sqlParser.SampleByBucketContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#sampleByBytes.
    def visitSampleByBytes(self, ctx:fugue_sqlParser.SampleByBytesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#identifierList.
    def visitIdentifierList(self, ctx:fugue_sqlParser.IdentifierListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#identifierSeq.
    def visitIdentifierSeq(self, ctx:fugue_sqlParser.IdentifierSeqContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#orderedIdentifierList.
    def visitOrderedIdentifierList(self, ctx:fugue_sqlParser.OrderedIdentifierListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#orderedIdentifier.
    def visitOrderedIdentifier(self, ctx:fugue_sqlParser.OrderedIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#identifierCommentList.
    def visitIdentifierCommentList(self, ctx:fugue_sqlParser.IdentifierCommentListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#identifierComment.
    def visitIdentifierComment(self, ctx:fugue_sqlParser.IdentifierCommentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#tableName.
    def visitTableName(self, ctx:fugue_sqlParser.TableNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#aliasedQuery.
    def visitAliasedQuery(self, ctx:fugue_sqlParser.AliasedQueryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#aliasedRelation.
    def visitAliasedRelation(self, ctx:fugue_sqlParser.AliasedRelationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#inlineTableDefault2.
    def visitInlineTableDefault2(self, ctx:fugue_sqlParser.InlineTableDefault2Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#tableValuedFunction.
    def visitTableValuedFunction(self, ctx:fugue_sqlParser.TableValuedFunctionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#inlineTable.
    def visitInlineTable(self, ctx:fugue_sqlParser.InlineTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#functionTable.
    def visitFunctionTable(self, ctx:fugue_sqlParser.FunctionTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#tableAlias.
    def visitTableAlias(self, ctx:fugue_sqlParser.TableAliasContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#rowFormatSerde.
    def visitRowFormatSerde(self, ctx:fugue_sqlParser.RowFormatSerdeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#rowFormatDelimited.
    def visitRowFormatDelimited(self, ctx:fugue_sqlParser.RowFormatDelimitedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#multipartIdentifierList.
    def visitMultipartIdentifierList(self, ctx:fugue_sqlParser.MultipartIdentifierListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#multipartIdentifier.
    def visitMultipartIdentifier(self, ctx:fugue_sqlParser.MultipartIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#tableIdentifier.
    def visitTableIdentifier(self, ctx:fugue_sqlParser.TableIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#functionIdentifier.
    def visitFunctionIdentifier(self, ctx:fugue_sqlParser.FunctionIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#namedExpression.
    def visitNamedExpression(self, ctx:fugue_sqlParser.NamedExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#namedExpressionSeq.
    def visitNamedExpressionSeq(self, ctx:fugue_sqlParser.NamedExpressionSeqContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#transformList.
    def visitTransformList(self, ctx:fugue_sqlParser.TransformListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#identityTransform.
    def visitIdentityTransform(self, ctx:fugue_sqlParser.IdentityTransformContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#applyTransform.
    def visitApplyTransform(self, ctx:fugue_sqlParser.ApplyTransformContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#transformArgument.
    def visitTransformArgument(self, ctx:fugue_sqlParser.TransformArgumentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#expression.
    def visitExpression(self, ctx:fugue_sqlParser.ExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#logicalNot.
    def visitLogicalNot(self, ctx:fugue_sqlParser.LogicalNotContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#predicated.
    def visitPredicated(self, ctx:fugue_sqlParser.PredicatedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#exists.
    def visitExists(self, ctx:fugue_sqlParser.ExistsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#logicalBinary.
    def visitLogicalBinary(self, ctx:fugue_sqlParser.LogicalBinaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#predicate.
    def visitPredicate(self, ctx:fugue_sqlParser.PredicateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#valueExpressionDefault.
    def visitValueExpressionDefault(self, ctx:fugue_sqlParser.ValueExpressionDefaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#comparison.
    def visitComparison(self, ctx:fugue_sqlParser.ComparisonContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#arithmeticBinary.
    def visitArithmeticBinary(self, ctx:fugue_sqlParser.ArithmeticBinaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#arithmeticUnary.
    def visitArithmeticUnary(self, ctx:fugue_sqlParser.ArithmeticUnaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#struct.
    def visitStruct(self, ctx:fugue_sqlParser.StructContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#dereference.
    def visitDereference(self, ctx:fugue_sqlParser.DereferenceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#simpleCase.
    def visitSimpleCase(self, ctx:fugue_sqlParser.SimpleCaseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#columnReference.
    def visitColumnReference(self, ctx:fugue_sqlParser.ColumnReferenceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#rowConstructor.
    def visitRowConstructor(self, ctx:fugue_sqlParser.RowConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#last.
    def visitLast(self, ctx:fugue_sqlParser.LastContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#star.
    def visitStar(self, ctx:fugue_sqlParser.StarContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#overlay.
    def visitOverlay(self, ctx:fugue_sqlParser.OverlayContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#subscript.
    def visitSubscript(self, ctx:fugue_sqlParser.SubscriptContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#subqueryExpression.
    def visitSubqueryExpression(self, ctx:fugue_sqlParser.SubqueryExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#substring.
    def visitSubstring(self, ctx:fugue_sqlParser.SubstringContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#currentDatetime.
    def visitCurrentDatetime(self, ctx:fugue_sqlParser.CurrentDatetimeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#cast.
    def visitCast(self, ctx:fugue_sqlParser.CastContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#constantDefault.
    def visitConstantDefault(self, ctx:fugue_sqlParser.ConstantDefaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#lambda.
    def visitLambda(self, ctx:fugue_sqlParser.LambdaContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#parenthesizedExpression.
    def visitParenthesizedExpression(self, ctx:fugue_sqlParser.ParenthesizedExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#extract.
    def visitExtract(self, ctx:fugue_sqlParser.ExtractContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#trim.
    def visitTrim(self, ctx:fugue_sqlParser.TrimContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#functionCall.
    def visitFunctionCall(self, ctx:fugue_sqlParser.FunctionCallContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#searchedCase.
    def visitSearchedCase(self, ctx:fugue_sqlParser.SearchedCaseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#position.
    def visitPosition(self, ctx:fugue_sqlParser.PositionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#first.
    def visitFirst(self, ctx:fugue_sqlParser.FirstContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#nullLiteral.
    def visitNullLiteral(self, ctx:fugue_sqlParser.NullLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#intervalLiteral.
    def visitIntervalLiteral(self, ctx:fugue_sqlParser.IntervalLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#typeConstructor.
    def visitTypeConstructor(self, ctx:fugue_sqlParser.TypeConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#numericLiteral.
    def visitNumericLiteral(self, ctx:fugue_sqlParser.NumericLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#booleanLiteral.
    def visitBooleanLiteral(self, ctx:fugue_sqlParser.BooleanLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#stringLiteral.
    def visitStringLiteral(self, ctx:fugue_sqlParser.StringLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#comparisonOperator.
    def visitComparisonOperator(self, ctx:fugue_sqlParser.ComparisonOperatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#comparisonEqualOperator.
    def visitComparisonEqualOperator(self, ctx:fugue_sqlParser.ComparisonEqualOperatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#arithmeticOperator.
    def visitArithmeticOperator(self, ctx:fugue_sqlParser.ArithmeticOperatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#predicateOperator.
    def visitPredicateOperator(self, ctx:fugue_sqlParser.PredicateOperatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#booleanValue.
    def visitBooleanValue(self, ctx:fugue_sqlParser.BooleanValueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#interval.
    def visitInterval(self, ctx:fugue_sqlParser.IntervalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#errorCapturingMultiUnitsInterval.
    def visitErrorCapturingMultiUnitsInterval(self, ctx:fugue_sqlParser.ErrorCapturingMultiUnitsIntervalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#multiUnitsInterval.
    def visitMultiUnitsInterval(self, ctx:fugue_sqlParser.MultiUnitsIntervalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#errorCapturingUnitToUnitInterval.
    def visitErrorCapturingUnitToUnitInterval(self, ctx:fugue_sqlParser.ErrorCapturingUnitToUnitIntervalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#unitToUnitInterval.
    def visitUnitToUnitInterval(self, ctx:fugue_sqlParser.UnitToUnitIntervalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#intervalValue.
    def visitIntervalValue(self, ctx:fugue_sqlParser.IntervalValueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#intervalUnit.
    def visitIntervalUnit(self, ctx:fugue_sqlParser.IntervalUnitContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#colPosition.
    def visitColPosition(self, ctx:fugue_sqlParser.ColPositionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#complexDataType.
    def visitComplexDataType(self, ctx:fugue_sqlParser.ComplexDataTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#primitiveDataType.
    def visitPrimitiveDataType(self, ctx:fugue_sqlParser.PrimitiveDataTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#qualifiedColTypeWithPositionList.
    def visitQualifiedColTypeWithPositionList(self, ctx:fugue_sqlParser.QualifiedColTypeWithPositionListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#qualifiedColTypeWithPosition.
    def visitQualifiedColTypeWithPosition(self, ctx:fugue_sqlParser.QualifiedColTypeWithPositionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#colTypeList.
    def visitColTypeList(self, ctx:fugue_sqlParser.ColTypeListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#colType.
    def visitColType(self, ctx:fugue_sqlParser.ColTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#complexColTypeList.
    def visitComplexColTypeList(self, ctx:fugue_sqlParser.ComplexColTypeListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#complexColType.
    def visitComplexColType(self, ctx:fugue_sqlParser.ComplexColTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#whenClause.
    def visitWhenClause(self, ctx:fugue_sqlParser.WhenClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#windowClause.
    def visitWindowClause(self, ctx:fugue_sqlParser.WindowClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#namedWindow.
    def visitNamedWindow(self, ctx:fugue_sqlParser.NamedWindowContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#windowRef.
    def visitWindowRef(self, ctx:fugue_sqlParser.WindowRefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#windowDef.
    def visitWindowDef(self, ctx:fugue_sqlParser.WindowDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#windowFrame.
    def visitWindowFrame(self, ctx:fugue_sqlParser.WindowFrameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#frameBound.
    def visitFrameBound(self, ctx:fugue_sqlParser.FrameBoundContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#qualifiedNameList.
    def visitQualifiedNameList(self, ctx:fugue_sqlParser.QualifiedNameListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#functionName.
    def visitFunctionName(self, ctx:fugue_sqlParser.FunctionNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#qualifiedName.
    def visitQualifiedName(self, ctx:fugue_sqlParser.QualifiedNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#errorCapturingIdentifier.
    def visitErrorCapturingIdentifier(self, ctx:fugue_sqlParser.ErrorCapturingIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#errorIdent.
    def visitErrorIdent(self, ctx:fugue_sqlParser.ErrorIdentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#realIdent.
    def visitRealIdent(self, ctx:fugue_sqlParser.RealIdentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#identifier.
    def visitIdentifier(self, ctx:fugue_sqlParser.IdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#unquotedIdentifier.
    def visitUnquotedIdentifier(self, ctx:fugue_sqlParser.UnquotedIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#quotedIdentifierAlternative.
    def visitQuotedIdentifierAlternative(self, ctx:fugue_sqlParser.QuotedIdentifierAlternativeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#quotedIdentifier.
    def visitQuotedIdentifier(self, ctx:fugue_sqlParser.QuotedIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#exponentLiteral.
    def visitExponentLiteral(self, ctx:fugue_sqlParser.ExponentLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#decimalLiteral.
    def visitDecimalLiteral(self, ctx:fugue_sqlParser.DecimalLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#legacyDecimalLiteral.
    def visitLegacyDecimalLiteral(self, ctx:fugue_sqlParser.LegacyDecimalLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#integerLiteral.
    def visitIntegerLiteral(self, ctx:fugue_sqlParser.IntegerLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#bigIntLiteral.
    def visitBigIntLiteral(self, ctx:fugue_sqlParser.BigIntLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#smallIntLiteral.
    def visitSmallIntLiteral(self, ctx:fugue_sqlParser.SmallIntLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#tinyIntLiteral.
    def visitTinyIntLiteral(self, ctx:fugue_sqlParser.TinyIntLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#doubleLiteral.
    def visitDoubleLiteral(self, ctx:fugue_sqlParser.DoubleLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#bigDecimalLiteral.
    def visitBigDecimalLiteral(self, ctx:fugue_sqlParser.BigDecimalLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#alterColumnAction.
    def visitAlterColumnAction(self, ctx:fugue_sqlParser.AlterColumnActionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#ansiNonReserved.
    def visitAnsiNonReserved(self, ctx:fugue_sqlParser.AnsiNonReservedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#strictNonReserved.
    def visitStrictNonReserved(self, ctx:fugue_sqlParser.StrictNonReservedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fugue_sqlParser#nonReserved.
    def visitNonReserved(self, ctx:fugue_sqlParser.NonReservedContext):
        return self.visitChildren(ctx)



del fugue_sqlParser


import random
from typing import Any, Iterable, List

import pyspark.sql as ps
from fugue_spark._utils.convert import to_schema, to_spark_schema
from pyspark import RDD
from pyspark.sql import SparkSession
from pyspark.sql.functions import lit

_PARTITION_DUMMY_KEY = "__partition_dummy_key__"


def hash_repartition(
    session: SparkSession, df: ps.DataFrame, num: int, cols: List[Any]
) -> ps.DataFrame:
    if num <= 0:
        if len(cols) == 0:
            return df
        return df.repartition(*cols)
    if num == 1:
        return _single_repartition(df)
    return df.repartition(num, *cols)


def rand_repartition(
    session: SparkSession, df: ps.DataFrame, num: int, cols: List[Any]
) -> ps.DataFrame:
    if len(cols) > 0 or num <= 1:
        return hash_repartition(session, df, num, cols)

    def _rand(rows: Iterable[Any], n: int) -> Iterable[Any]:  # pragma: no cover
        for row in rows:
            yield (random.randint(0, n - 1), row)

    rdd = (
        df.rdd.mapPartitions(lambda r: _rand(r, num))
        .partitionBy(num, lambda k: k)
        .mapPartitions(_to_rows)
    )
    return session.createDataFrame(rdd, df.schema)


def even_repartition(
    session: SparkSession, df: ps.DataFrame, num: int, cols: List[Any]
) -> ps.DataFrame:
    if num == 1:
        return _single_repartition(df)
    if len(cols) == 0:
        if num == 0:
            return df
        rdd = (
            _zipWithIndex(df.rdd).partitionBy(num, lambda k: k).mapPartitions(_to_rows)
        )
        return session.createDataFrame(rdd, df.schema)
    else:
        keys = df.select(*cols).distinct()
        krdd = _zipWithIndex(keys.rdd, True)
        new_schema = to_spark_schema(
            to_schema(df.schema).extract(cols) + f"{_PARTITION_DUMMY_KEY}:long"
        )
        idx = session.createDataFrame(krdd, new_schema)
        if num <= 0:
            idx = idx.persist()
            num = idx.count()
        idf = (
            df.alias("df")
            .join(idx.alias("idx"), on=cols, how="inner")
            .select(_PARTITION_DUMMY_KEY, *["df." + x for x in df.schema.names])
        )

        def _to_kv(rows: Iterable[Any]) -> Iterable[Any]:  # pragma: no cover
            for row in rows:
                yield (row[0], row[1:])

        rdd = (
            idf.rdd.mapPartitions(_to_kv)
            .partitionBy(num, lambda k: k)
            .mapPartitions(_to_rows)
        )
        return session.createDataFrame(rdd, df.schema)


def _single_repartition(df: ps.DataFrame) -> ps.DataFrame:
    return (
        df.withColumn(_PARTITION_DUMMY_KEY, lit(0))
        .repartition(_PARTITION_DUMMY_KEY)
        .drop(_PARTITION_DUMMY_KEY)
    )


def _to_rows(rdd: Iterable[Any]) -> Iterable[Any]:  # pragma: no cover
    for item in rdd:
        yield item[1]


def _to_rows_with_key(rdd: Iterable[Any]) -> Iterable[Any]:  # pragma: no cover
    for item in rdd:
        yield list(item[1]) + [item[0]]


def _zipWithIndex(rdd: RDD, to_rows: bool = False) -> RDD:
    """
    Modified from
    https://github.com/davies/spark/blob/cebe5bfe263baf3349353f1473f097396821514a/python/pyspark/rdd.py

    """
    starts = [0]
    if rdd.getNumPartitions() > 1:
        nums = rdd.mapPartitions(lambda it: [sum(1 for i in it)]).collect()
        for i in range(len(nums) - 1):
            starts.append(starts[-1] + nums[i])

    def func1(k, it):  # pragma: no cover
        for i, v in enumerate(it, starts[k]):
            yield i, v

    def func2(k, it):  # pragma: no cover
        for i, v in enumerate(it, starts[k]):
            yield list(v) + [i]

    if not to_rows:
        return rdd.mapPartitionsWithIndex(func1)
    else:
        return rdd.mapPartitionsWithIndex(func2)



# flake8: noqa
from fugue.dataframe.array_dataframe import ArrayDataFrame
from fugue.dataframe.arrow_dataframe import ArrowDataFrame
from fugue.dataframe.dataframe import DataFrame, LocalBoundedDataFrame, LocalDataFrame
from fugue.dataframe.dataframes import DataFrames
from fugue.dataframe.iterable_dataframe import IterableDataFrame
from fugue.dataframe.pandas_dataframe import PandasDataFrame
from fugue.dataframe.utils import to_local_bounded_df, to_local_df



from fugue_spark._utils.convert import (to_cast_expression, to_schema,
                                       to_select_expression, to_spark_schema)
from pytest import raises


def test_schema_conversion(spark_session):
    def test(expr):
        assert to_schema(to_spark_schema(expr)) == expr

    test("a:int,b:long,c:[int],d:datetime,e:date,f:decimal(3,4),g:str")
    test("a:{a:[int],b:[str]}")
    # test("a:[{a:int}]") TODO: this is not supported by spark, should we support?
    s = to_spark_schema(to_spark_schema("a:int"))
    assert to_spark_schema(s) is s

    df = spark_session.createDataFrame([[1]], schema=to_spark_schema("a:int"))
    assert to_schema(to_spark_schema(df)) == "a:int"
    assert to_schema(df) == "a:int"
    assert to_schema(dict(a=str)) == "a:str"


def test_to_cast_expression():
    # length mismatch
    raises(ValueError, lambda: to_cast_expression(
        "a:int,b:int", "a:int", False))
    assert (False, ["a", "b"]) == to_cast_expression(
        "a:int,b:int", "a:int,b:int", False)
    assert (False, ["a", "b"]) == to_cast_expression(
        "a:int,b:int", "a:int,b:int", True)
    assert (True, ["aa AS a", "b"]) == to_cast_expression(
        "aa:int,b:int", "a:int,b:int", True)
    raises(ValueError, lambda: to_cast_expression(
        "aa:int,b:int", "a:int,b:int", False))
    assert (True, ["CAST(a AS int) a", "b"]) == to_cast_expression(
        "a:long,b:int", "a:int,b:int", True)
    assert (True, ["CAST(aa AS int) a", "b"]) == to_cast_expression(
        "aa:long,b:int", "a:int,b:int", True)


def test_to_select_expression():
    assert to_select_expression("a:int,b:str", ["b", "a"]) == ["b", "a"]
    raises(KeyError, lambda: to_select_expression("a:int,b:str", ["b", "x"]))
    assert to_select_expression("a:int,b:str", "b:str,a:int") == ["b", "a"]
    assert to_select_expression("a:int,b:str", "b:str,a:long") == [
        "b", "CAST(a AS bigint) a"]
    assert to_select_expression("a:int,b:double,c:float", "a:str,b:str,c:str") == [
        "CAST(a AS string) a",
        "CAST(IF(isnan(b), NULL, b) AS string) b",
        "CAST(IF(isnan(c), NULL, c) AS string) c"]
    raises(KeyError, lambda: to_select_expression("a:int,b:str",  "b:str,x:int"))



import os

import pandas as pd
from fugue.dataframe import DataFrame
from fugue.dataframe.array_dataframe import ArrayDataFrame
from fugue.dataframe.utils import _df_eq
from fugue.execution.native_execution_engine import NativeExecutionEngine
from fugue.workflow.workflow import WorkflowDataFrame
from fugue.extensions._builtins.outputters import Show
from pytest import raises

from fugue_sql.exceptions import FugueSQLError, FugueSQLSyntaxError
from fugue_sql.workflow import FugueSQLWorkflow


def test_workflow_conf():
    dag = FugueSQLWorkflow(
        NativeExecutionEngine({"x": 10, "fugue.sql.compile.simple_assign": "false"})
    )
    assert 10 == dag.conf.get_or_throw("x", int)
    assert not dag.conf.get_or_throw("fugue.sql.compile.simple_assign", bool)
    assert not dag.conf.get_or_throw("fugue.sql.compile.ignore_case", bool)


def test_conf_override():
    with raises(FugueSQLSyntaxError):
        FugueSQLWorkflow()("create [[0]] schema a:int")
    with FugueSQLWorkflow(
        NativeExecutionEngine({"fugue.sql.compile.ignore_case": "true"})
    ) as dag:
        a = dag.df([[0], [1]], "a:int")
        dag(
            """
        b = create [[0],[1]] schema a:int
        output a,b using assert_eq"""
        )


def test_show():
    class _CustomShow(object):
        def __init__(self):
            self.called = False

        def show(self, schema, head_rows, title, rows, count):
            print(schema, head_rows)
            print(title, rows, count)
            self.called = True

    cs = _CustomShow()
    with FugueSQLWorkflow() as dag:
        dag(
            """
        a = CREATE[[0], [1]] SCHEMA a: int
        b = CREATE[[0], [1]] SCHEMA a: int
        PRINT a, b ROWS 10 ROWCOUNT TITLE "abc"
        PRINT a, b
        """
        )
    assert not cs.called
    Show.set_hook(cs.show)
    with FugueSQLWorkflow() as dag:
        dag(
            """
        a = CREATE[[0], [1]] SCHEMA a: int
        b = CREATE[[0], [1]] SCHEMA a: int
        PRINT a, b ROWS 10 ROWCOUNT TITLE "abc"
        PRINT a, b
        """
        )

    assert cs.called


def test_jinja_keyword_in_sql():
    with FugueSQLWorkflow() as dag:
        dag(
            """
        CREATE [["{%'{%'"]] SCHEMA a:str
        SELECT * WHERE a LIKE '{%'
        PRINT
        """
        )


def test_use_df(tmpdir):
    # df generated inside dag
    with FugueSQLWorkflow() as dag:
        a = dag.df([[0], [1]], "a:int")
        dag(
            """
        b=CREATE[[0], [1]] SCHEMA a: int
        OUTPUT a, b USING assert_eq
        """
        )
        dag["b"].assert_eq(a)

    # external non-workflowdataframe
    arr = ArrayDataFrame([[0], [1]], "a:int")
    with FugueSQLWorkflow() as dag:
        dag(
            """
        b=CREATE[[0], [1]] SCHEMA a: int
        OUTPUT a, b USING assert_eq
        """,
            a=arr,
        )
        dag["b"].assert_eq(dag.df([[0], [1]], "a:int"))

    # from yield
    engine = NativeExecutionEngine(
        conf={"fugue.workflow.checkpoint.path": os.path.join(tmpdir, "ck")}
    )
    with FugueSQLWorkflow(engine) as dag:
        dag("CREATE[[0], [1]] SCHEMA a: int YIELD AS b")
        res = dag.yields["b"]

    with FugueSQLWorkflow(engine) as dag:
        dag(
            """
        b=CREATE[[0], [1]] SCHEMA a: int
        OUTPUT a, b USING assert_eq
        """,
            a=res,
        )


def test_use_param():
    with FugueSQLWorkflow() as dag:
        a = dag.df([[0], [1]], "a:int")
        x = 0
        dag(
            """
        b=CREATE[[{{x}}], [{{y}}]] SCHEMA a: int
        OUTPUT a, b USING assert_eq
        """,
            y=1,
        )


def test_multiple_sql():
    with FugueSQLWorkflow() as dag:
        a = dag.df([[0], [1]], "a:int")
        dag("b = CREATE [[0],[1]] SCHEMA a:int")
        dag("OUTPUT a,b USING assert_eq")
        assert dag["a"] is a
        assert isinstance(dag["b"], WorkflowDataFrame)


def test_multiple_sql_with_reset():
    with FugueSQLWorkflow() as dag:
        a = dag.df([[0], [1]], "a:int")
        dag("b = CREATE [[0],[1]] SCHEMA a:int")
        a = dag.df([[0], [2]], "a:int")
        b = dag.df([[0], [2]], "a:int")
        dag(
            """
        OUTPUT a, b USING assert_eq
        OUTPUT a, (CREATE[[0], [2]] SCHEMA a: int) USING assert_eq
        """
        )


def test_multiple_blocks():
    with FugueSQLWorkflow() as dag:
        a = dag.df([[0], [1]], "a:int")
        c = 1
        dag(
            """
        OUTPUT a, (CREATE[[0], [1]] SCHEMA a: int) USING assert_eq
        """
        )
    # dataframe can't pass to another workflow
    with FugueSQLWorkflow() as dag:
        assert "a" in locals()
        with raises(FugueSQLError):
            dag(
                """
            OUTPUT a, (CREATE[[0], [1]] SCHEMA a: int) USING assert_eq
            """
            )
    # other local variables are fine
    with FugueSQLWorkflow() as dag:
        a = dag.df([[0], [1]], "a:int")
        dag(
            """
        OUTPUT a, (CREATE[[0], [{{c}}]] SCHEMA a: int) USING assert_eq
        """
        )


def test_function_calls():
    with FugueSQLWorkflow() as dag:
        a = dag.df([[0], [1]], "a:int")
        _eq(dag, a)


def test_local_instance_as_extension():
    class _Mock(object):
        # schema: *
        def t(self, df: pd.DataFrame) -> pd.DataFrame:
            return df

        def test(self):
            with FugueSQLWorkflow() as dag:
                dag(
                    """
                a = CREATE [[0],[1]] SCHEMA a:int
                b = TRANSFORM USING self.t
                OUTPUT a,b USING assert_eq
                """
                )

    m = _Mock()
    m.test()
    with FugueSQLWorkflow() as dag:
        dag(
            """
        a = CREATE [[0],[1]] SCHEMA a:int
        b = TRANSFORM USING m.t
        OUTPUT a,b USING assert_eq
        """
        )


def _eq(dag, a):
    dag(
        """
    OUTPUT a, (CREATE[[0], [1]] SCHEMA a: int) USING assert_eq
    """
    )


def assert_eq(df1: DataFrame, df2: DataFrame) -> None:
    _df_eq(df1, df2, throw=True)



import copy
from typing import Any, Callable, Dict, List, Optional, no_type_check

from fugue._utils.interfaceless import FunctionWrapper
from fugue.dataframe import DataFrames
from fugue.exceptions import FugueInterfacelessError
from fugue.extensions._utils import (
    parse_validation_rules_from_comment,
    to_validation_rules,
)
from fugue.extensions.outputter.outputter import Outputter
from triad.utils.convert import get_caller_global_local_vars, to_function, to_instance
from triad.utils.hash import to_uuid


def outputter(**validation_rules: Any) -> Callable[[Any], "_FuncAsOutputter"]:
    """Decorator for outputters

    Please read :ref:`Outputter Tutorial <tutorial:/tutorials/outputter.ipynb>`
    """

    def deco(func: Callable) -> "_FuncAsOutputter":
        return _FuncAsOutputter.from_func(
            func, validation_rules=to_validation_rules(validation_rules)
        )

    return deco


def _to_outputter(
    obj: Any,
    global_vars: Optional[Dict[str, Any]] = None,
    local_vars: Optional[Dict[str, Any]] = None,
    validation_rules: Optional[Dict[str, Any]] = None,
) -> Outputter:
    global_vars, local_vars = get_caller_global_local_vars(global_vars, local_vars)
    exp: Optional[Exception] = None
    if validation_rules is None:
        validation_rules = {}
    try:
        return copy.copy(
            to_instance(obj, Outputter, global_vars=global_vars, local_vars=local_vars)
        )
    except Exception as e:
        exp = e
    try:
        f = to_function(obj, global_vars=global_vars, local_vars=local_vars)
        # this is for string expression of function with decorator
        if isinstance(f, Outputter):
            return copy.copy(f)
        # this is for functions without decorator
        return _FuncAsOutputter.from_func(f, validation_rules=validation_rules)
    except Exception as e:
        exp = e
    raise FugueInterfacelessError(f"{obj} is not a valid outputter", exp)


class _FuncAsOutputter(Outputter):
    @property
    def validation_rules(self) -> Dict[str, Any]:
        return self._validation_rules  # type: ignore

    @no_type_check
    def process(self, dfs: DataFrames) -> None:
        args: List[Any] = []
        kwargs: Dict[str, Any] = {}
        if self._need_engine:
            args.append(self.execution_engine)
        if self._use_dfs:
            args.append(dfs)
        else:
            if not dfs.has_key:
                args += dfs.values()
            else:
                kwargs.update(dfs)
        kwargs.update(self.params)
        return self._wrapper.run(args=args, kwargs=kwargs)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self._wrapper(*args, **kwargs)  # type: ignore

    @no_type_check
    def __uuid__(self) -> str:
        return to_uuid(self._wrapper, self._need_engine, self._use_dfs)

    @no_type_check
    @staticmethod
    def from_func(
        func: Callable, validation_rules: Dict[str, Any]
    ) -> "_FuncAsOutputter":
        validation_rules.update(parse_validation_rules_from_comment(func))
        tr = _FuncAsOutputter()
        tr._wrapper = FunctionWrapper(  # type: ignore
            func, "^e?(c|[dlsp]+)x*z?$", "^n$"
        )
        tr._need_engine = tr._wrapper.input_code.startswith("e")
        tr._use_dfs = "c" in tr._wrapper.input_code
        tr._validation_rules = validation_rules
        return tr



from fugue_sql._utils import fill_sql_template


def test_fill_sql_template():
    data = {"a": 1, "b": "x"}
    assert "a=select " == fill_sql_template("a=select ", data)
    assert "a=select * from b like '{%'" == fill_sql_template(
        "a=select * from b like '{%'", data
    )
    assert 'a=select * from b like "{%"' == fill_sql_template(
        'a=select * from b like "{%"', data
    )
    assert "a=select * from b like \"{%'{%'\"" == fill_sql_template(
        "a=select * from b like \"{%'{%'\"", data
    )
    assert 'a=select * from b like "%}"' == fill_sql_template(
        'a=select * from b like "%}"', data
    )
    assert (
        "a=select * from b like \"{% for env in ('dev', 'prod') %}\""
        == fill_sql_template(
            "a=select * from b like \"{% for env in ('dev', 'prod') %}\"", data
        )
    )
    assert (
        "a=select * from b like \"{% for env in ('dev', 'prod') %}{% endfor %}\""
        == fill_sql_template(
            "a=select * from b like \"{% for env in ('dev', 'prod') %}{% endfor %}\"",
            data,
        )
    )

    assert "a=select * from b like '{%'" == fill_sql_template(
        "a=select * from b like '{%'", data
    )
    assert "a=select * from b like '%}'" == fill_sql_template(
        "a=select * from b like '%}'", data
    )
    assert (
        "a=select * from b like '{% for env in ('dev', 'prod') %}'"
        == fill_sql_template(
            "a=select * from b like '{% for env in ('dev', 'prod') %}'", data
        )
    )
    assert (
        "a=select * from b like '{% for env in ('dev', 'prod') %}{% endfor %}'"
        == fill_sql_template(
            "a=select * from b like '{% for env in ('dev', 'prod') %}{% endfor %}'",
            data,
        )
    )

    assert "1x1" == fill_sql_template("{{a}}{{b}}{{a}}", data)
    assert "" == fill_sql_template("", data)
    assert "%s" == fill_sql_template("%s", data)
    assert "%%s" == fill_sql_template("%%s", data)
    assert "1%%sx1" == fill_sql_template("{{a}}%%s{{b}}{{a}}", data)

    assert "1" == fill_sql_template("{{a}}", {"a": 1, "self": 2})



import json
from datetime import datetime
from typing import Any

import numpy as np
import pandas as pd
from fugue.dataframe import ArrowDataFrame, PandasDataFrame
from fugue.dataframe.utils import _df_eq as df_eq
from fugue.exceptions import FugueDataFrameInitError
from fugue_test.dataframe_suite import DataFrameTests
from pytest import raises
from triad.collections.schema import Schema, SchemaError
from triad.exceptions import InvalidOperationError


class ArrowDataFrameTests(DataFrameTests.Tests):
    def df(
        self, data: Any = None, schema: Any = None, metadata: Any = None
    ) -> ArrowDataFrame:
        return ArrowDataFrame(data, schema, metadata)


def test_init():
    df = ArrowDataFrame(schema="a:str,b:int")
    assert df.empty
    assert df.schema == "a:str,b:int"
    assert df.is_bounded

    data = [["a", "1"], ["b", "2"]]
    df = ArrowDataFrame(data, "a:str,b:str")
    assert [["a", "1"], ["b", "2"]] == df.as_array(type_safe=True)
    data = [["a", 1], ["b", 2]]
    df = ArrowDataFrame(data, "a:str,b:int")
    assert [["a", 1.0], ["b", 2.0]] == df.as_array(type_safe=True)
    df = ArrowDataFrame(data, "a:str,b:double")
    assert [["a", 1.0], ["b", 2.0]] == df.as_array(type_safe=True)

    ddf = ArrowDataFrame(df.native)
    assert [["a", 1.0], ["b", 2.0]] == ddf.as_array(type_safe=True)

    df = ArrowDataFrame(df.as_pandas(), "a:str,b:double")
    assert [["a", 1.0], ["b", 2.0]] == df.as_array(type_safe=True)
    df = ArrowDataFrame(df.as_pandas()["b"])
    assert [[1.0], [2.0]] == df.as_array(type_safe=True)

    df = ArrowDataFrame([], "x:str,y:double")
    assert df.empty
    assert df.is_local
    assert df.is_bounded

    raises(FugueDataFrameInitError, lambda: ArrowDataFrame(123))



import os

from fugue.execution import NativeExecutionEngine
from fugue.execution.execution_engine import _get_file_threshold
from fugue_test.builtin_suite import BuiltInTests
from fugue_test.execution_suite import ExecutionEngineTests


class NativeExecutionEngineTests(ExecutionEngineTests.Tests):
    def make_engine(self):
        e = NativeExecutionEngine(dict(test=True))
        return e

    def test_map_with_dict_col(self):
        # TODO: add back
        return


class NativeExecutionEngineBuiltInTests(BuiltInTests.Tests):
    def make_engine(self):
        e = NativeExecutionEngine(dict(test=True))
        return e


def test_get_file_threshold():
    assert -1 == _get_file_threshold(None)
    assert -2 == _get_file_threshold(-2)
    assert 1024 == _get_file_threshold("1k")



import datetime
import os
import pickle
from typing import Any, Dict, Iterable, List
from unittest import TestCase

import numpy as np
import pandas as pd
import pytest
from fugue import FileSystem, Schema
from fugue.collections.partition import PartitionSpec
from fugue.dataframe import DataFrame, DataFrames, LocalDataFrame, PandasDataFrame
from fugue.dataframe.array_dataframe import ArrayDataFrame
from fugue.dataframe.utils import _df_eq as df_eq
from fugue.exceptions import (
    FugueWorkflowCompileError,
    FugueWorkflowCompileValidationError,
    FugueWorkflowError,
    FugueWorkflowRuntimeValidationError,
)
from fugue.execution import ExecutionEngine
from fugue.execution.native_execution_engine import SqliteEngine
from fugue.extensions.outputter import Outputter, outputter
from fugue.extensions.processor import Processor, processor
from fugue.extensions.transformer import (
    CoTransformer,
    OutputCoTransformer,
    OutputTransformer,
    Transformer,
    cotransformer,
    output_cotransformer,
    output_transformer,
    transformer,
)
from fugue.workflow.workflow import FugueWorkflow, _FugueInteractiveWorkflow
from pytest import raises


class BuiltInTests(object):
    """Workflow level general test suite. It is a more general end to end
    test suite than :class:`~fugue_test.execution_suite.ExecutionEngineTests`.
    Any new :class:`~fugue.execution.execution_engine.ExecutionEngine`
    should also pass this test suite.

    Whenever you add method to FugueWorkflow and WorkflowDataFrame, you should
    add correspondent tests here
    """

    class Tests(TestCase):
        @classmethod
        def setUpClass(cls):
            cls._engine = cls.make_engine(cls)

        @property
        def engine(self) -> ExecutionEngine:
            return self._engine  # type: ignore

        @classmethod
        def tearDownClass(cls):
            cls._engine.stop()

        def make_engine(self) -> ExecutionEngine:  # pragma: no cover
            raise NotImplementedError

        def dag(self) -> FugueWorkflow:
            return FugueWorkflow(self.engine)

        def test_workflows(self):
            a = FugueWorkflow().df([[0]], "a:int")
            df_eq(a.compute(self.engine), [[0]], "a:int")

            a = _FugueInteractiveWorkflow(self.engine).df([[0]], "a:int").persist()
            df_eq(a.result, [[0]], "a:int")

        def test_create_show(self):
            with self.dag() as dag:
                dag.df([[0]], "a:int").persist().partition(num=2).show()
                dag.df(dag.df([[0]], "a:int")).persist().broadcast().show(title="t")

        def test_checkpoint(self):
            with raises(FugueWorkflowError):
                with self.dag() as dag:
                    dag.df([[0]], "a:int").checkpoint()

            self.engine.conf["fugue.workflow.checkpoint.path"] = os.path.join(
                self.tmpdir, "ck"
            )
            with self.dag() as dag:
                a = dag.df([[0]], "a:int").checkpoint()
                dag.df([[0]], "a:int").assert_eq(a)

        def test_deterministic_checkpoint(self):
            self.engine.conf["fugue.workflow.checkpoint.path"] = os.path.join(
                self.tmpdir, "ck"
            )
            temp_file = os.path.join(self.tmpdir, "t.parquet")

            def mock_create(dummy: int = 1) -> pd.DataFrame:
                return pd.DataFrame(np.random.rand(3, 2), columns=["a", "b"])

            # base case without checkpoint, two runs should generate different result
            with self.dag() as dag:
                a = dag.create(mock_create)
                a.save(temp_file)
            with self.dag() as dag:
                a = dag.create(mock_create)
                b = dag.load(temp_file)
                b.assert_not_eq(a)

            # strong checkpoint is not cross execution, so result should be different
            with self.dag() as dag:
                a = dag.create(mock_create).strong_checkpoint()
                a.save(temp_file)
            with self.dag() as dag:
                a = dag.create(mock_create).strong_checkpoint()
                b = dag.load(temp_file)
                b.assert_not_eq(a)

            # with deterministic checkpoint, two runs should generate the same result
            with self.dag() as dag:
                dag.create(
                    mock_create, params=dict(dummy=2)
                )  # this should not affect a because it's not a dependency
                a = dag.create(mock_create).deterministic_checkpoint()
                id1 = a.spec_uuid()
                a.save(temp_file)
            with self.dag() as dag:
                a = dag.create(mock_create).deterministic_checkpoint()
                b = dag.load(temp_file)
                b.assert_eq(a)
            with self.dag() as dag:
                # checkpoint specs itself doesn't change determinism
                # of the previous steps
                a = dag.create(mock_create).deterministic_checkpoint(
                    partition=PartitionSpec(num=2)
                )
                id2 = a.spec_uuid()
                b = dag.load(temp_file)
                b.assert_eq(a)
            with self.dag() as dag:
                # dependency change will affect determinism
                a = dag.create(
                    mock_create, params={"dummy": 2}
                ).deterministic_checkpoint()
                id3 = a.spec_uuid()
                b = dag.load(temp_file)
                b.assert_not_eq(a)

            assert (
                id1 == id2
            )  # different types of checkpoint doesn't change dag determinism
            assert id1 != id3

        def test_deterministic_checkpoint_complex_dag(self):
            self.engine.conf["fugue.workflow.checkpoint.path"] = os.path.join(
                self.tmpdir, "ck"
            )
            temp_file = os.path.join(self.tmpdir, "t.parquet")

            def mock_create(dummy: int = 1) -> pd.DataFrame:
                return pd.DataFrame(np.random.rand(3, 2), columns=["a", "b"])

            # base case without checkpoint, two runs should generate different result
            with self.dag() as dag:
                a = dag.create(mock_create).drop(["a"])
                b = dag.create(mock_create).drop(["a"])
                c = a.union(b)
                c.save(temp_file)
            with self.dag() as dag:
                a = dag.create(mock_create).drop(["a"])
                b = dag.create(mock_create).drop(["a"])
                c = a.union(b)
                d = dag.load(temp_file)
                d.assert_not_eq(c)

            # strong checkpoint is not cross execution, so result should be different
            with self.dag() as dag:
                a = dag.create(mock_create).drop(["a"])
                b = dag.create(mock_create).drop(["a"])
                c = a.union(b).strong_checkpoint()
                c.save(temp_file)
            with self.dag() as dag:
                a = dag.create(mock_create).drop(["a"])
                b = dag.create(mock_create).drop(["a"])
                c = a.union(b).strong_checkpoint()
                d = dag.load(temp_file)
                d.assert_not_eq(c)

            # with deterministic checkpoint, two runs should generate the same result
            with self.dag() as dag:
                dag.create(
                    mock_create, params=dict(dummy=2)
                )  # this should not affect a because it's not a dependency
                a = dag.create(mock_create).drop(["a"])
                b = dag.create(mock_create).drop(["a"])
                c = a.union(b)
                c.deterministic_checkpoint()
                c.save(temp_file)
            with self.dag() as dag:
                a = dag.create(mock_create).drop(["a"])
                b = dag.create(mock_create).drop(["a"])
                c = a.union(b).deterministic_checkpoint()
                d = dag.load(temp_file)
                d.assert_eq(c)
            with self.dag() as dag:
                # checkpoint specs itself doesn't change determinism
                # of the previous steps
                a = dag.create(mock_create).drop(["a"])
                b = dag.create(mock_create).drop(["a"])
                c = a.union(b).deterministic_checkpoint(partition=PartitionSpec(num=2))
                d = dag.load(temp_file)
                d.assert_eq(c)
            with self.dag() as dag:
                # dependency change will affect determinism
                a = dag.create(mock_create, params={"dummy": 2}).drop(["a"])
                b = dag.create(mock_create).drop(["a"])
                c = a.union(b).deterministic_checkpoint(partition=PartitionSpec(num=2))
                d = dag.load(temp_file)
                d.assert_not_eq(c)

        def test_yield(self):
            self.engine.conf["fugue.workflow.checkpoint.path"] = os.path.join(
                self.tmpdir, "ck"
            )

            with raises(FugueWorkflowCompileError):
                self.dag().df([[0]], "a:int").checkpoint().yield_as("x")

            with raises(FugueWorkflowCompileError):
                self.dag().df([[0]], "a:int").persist().yield_as("x")

            def run_test(deterministic):
                dag1 = self.dag()
                df = dag1.df([[0]], "a:int")
                if deterministic:
                    df = df.deterministic_checkpoint()
                df.yield_as("x")
                id1 = dag1.spec_uuid()
                dag2 = self.dag()
                dag2.df([[0]], "a:int").assert_eq(dag2.df(dag1.yields["x"]))
                id2 = dag2.spec_uuid()
                dag1.run()
                dag2.run()
                return id1, id2

            id1, id2 = run_test(False)
            id3, id4 = run_test(False)
            assert id1 == id3
            assert id2 != id4  # non deterministic yield (direct yield)

            id1, id2 = run_test(True)
            id3, id4 = run_test(True)
            assert id1 == id3
            assert id2 == id4  # deterministic yield (yield deterministic checkpoint)

        def test_create_process_output(self):
            with self.dag() as dag:
                a = dag.create(mock_creator, params=dict(p=2))
                a.assert_eq(ArrayDataFrame([[2]], "a:int"))
                b = dag.process(a, a, using=mock_processor)
                b.assert_eq(ArrayDataFrame([[2]], "a:int"))
                b = dag.process(dict(df1=a, df2=a), using=mock_processor)
                b.assert_eq(ArrayDataFrame([[2]], "a:int"))
                dag.output(a, b, using=mock_outputter)
                b2 = dag.process(a, a, a, using=mock_processor2)
                b2.assert_eq(ArrayDataFrame([[3]], "a:int"))
                b2 = dag.process(a, a, a, using=MockProcessor3)
                b2.assert_eq(ArrayDataFrame([[3]], "a:int"))
                a.process(mock_processor2).assert_eq(ArrayDataFrame([[1]], "a:int"))
                a.output(mock_outputter2)
                dag.output(dict(df=a), using=mock_outputter2)
                a.partition(num=3).output(MockOutputter3)
                dag.output(dict(aa=a, bb=b), using=MockOutputter4)

        def test_zip(self):
            with self.dag() as dag:
                a = dag.df([[1, 2], [2, 3], [2, 5]], "a:int,b:int")
                b = dag.df([[1, 3]], "a:int,c:int")
                c1 = a.zip(b)
                c2 = dag.zip(a, b)
                c1.assert_eq(c2, check_metadata=False)

                a = dag.df([[1, 2], [2, 3], [2, 5]], "a:int,b:int")
                b = dag.df([[1, 3]], "a:int,c:int")
                c1 = a.zip(b, how="left_outer", partition=dict(presort="b DESC, c ASC"))
                c2 = dag.zip(
                    a, b, how="left_outer", partition=dict(presort="b DESC, c ASC")
                )
                c1.assert_eq(c2, check_metadata=False)

                a = dag.df([[1, 2, 0], [1, 3, 1]], "a:int,b:int,c:int")
                b = dag.df([[1, 2, 1], [1, 3, 2]], "a:int,b:int,d:int")
                c = dag.df([[1, 4]], "a:int,e:int")
                e = dag.df([[1, 2], [1, 3]], "a:int,b:int")
                dag.zip(a, b, c)[["a", "b"]].assert_eq(e, check_metadata=False)

        def test_transform(self):
            with self.dag() as dag:
                a = dag.df([[1, 2], [3, 4]], "a:double,b:int", dict(x=1))
                c = a.transform(mock_tf0)
                dag.df([[1, 2, 1], [3, 4, 1]], "a:double,b:int,p:int").assert_eq(c)

                c = a.transform(mock_tf0, params=dict(col="x"))
                dag.df([[1, 2, 1], [3, 4, 1]], "a:double,b:int,x:int").assert_eq(c)

                a = dag.df(
                    [[1, 2], [None, 1], [3, 4], [None, 4]], "a:double,b:int", dict(x=1)
                )
                c = a.transform(mock_tf0, params=dict(p="10"))
                dag.df(
                    [[1, 2, 10], [None, 1, 10], [3, 4, 10], [None, 4, 10]],
                    "a:double,b:int,p:int",
                ).assert_eq(c)

        def test_local_instance_as_extension(self):
            class _Mock(object):
                # schema: *
                def t1(self, df: pd.DataFrame) -> pd.DataFrame:
                    return df

                def t2(self, df: pd.DataFrame) -> pd.DataFrame:
                    return df

                def test(self):
                    with FugueWorkflow() as dag_:
                        a = dag_.df([[0], [1]], "a:int")
                        b = a.transform(self.t1)
                        b.assert_eq(a)

            m = _Mock()
            m.test()
            with self.dag() as dag:
                a = dag.df([[0], [1]], "a:int")
                b = a.transform(m.t1).transform(m.t2, schema="*")
                b.assert_eq(a)

        def test_transform_binary(self):
            with self.dag() as dag:
                a = dag.df([[1, pickle.dumps([0, "a"])]], "a:int,b:bytes")
                c = a.transform(mock_tf3)
                b = dag.df([[1, pickle.dumps([1, "ax"])]], "a:int,b:bytes")
                b.assert_eq(c, check_order=True)

        def test_transform_by(self):
            with self.dag() as dag:
                a = dag.df(
                    [[1, 2], [None, 1], [3, 4], [None, 4]], "a:double,b:int", dict(x=1)
                )
                c = a.transform(MockTransform1, pre_partition={"by": ["a"]})
                dag.df(
                    [[None, 1, 2, 1], [None, 4, 2, 1], [1, 2, 1, 1], [3, 4, 1, 1]],
                    "a:double,b:int,ct:int,p:int",
                ).assert_eq(c)

                c = a.transform(
                    mock_tf1, pre_partition={"by": ["a"], "presort": "b DESC"}
                )
                dag.df(
                    [[None, 4, 2, 1], [None, 1, 2, 1], [1, 2, 1, 1], [3, 4, 1, 1]],
                    "a:double,b:int,ct:int,p:int",
                ).assert_eq(c)

                c = a.transform(
                    mock_tf2_except,
                    schema="*",
                    pre_partition={"by": ["a"], "presort": "b DESC"},
                    ignore_errors=[NotImplementedError],
                )
                dag.df([[1, 2], [3, 4]], "a:double,b:int").assert_eq(c)

                c = a.partition(by=["a"], presort="b DESC").transform(
                    mock_tf2_except, schema="*", ignore_errors=[NotImplementedError]
                )
                dag.df([[1, 2], [3, 4]], "a:double,b:int").assert_eq(c)

        def test_cotransform(self):
            with self.dag() as dag:
                a = dag.df([[1, 2], [1, 3], [2, 1]], "a:int,b:int", dict(x=1))
                b = dag.df([[1, 2], [3, 4]], "a:int,c:int", dict(x=1))
                c = dag.transform(a.zip(b), using=MockCoTransform1)
                e = dag.df([[1, 2, 1, 1]], "a:int,ct1:int,ct2:int,p:int")
                e.assert_eq(c)
                c = dag.transform(a.zip(b), using=MockCoTransform1, params=dict(p=10))
                e = dag.df([[1, 2, 1, 10]], "a:int,ct1:int,ct2:int,p:int")
                e.assert_eq(c)
                # with deco
                c = dag.transform(a.zip(b), using=mock_co_tf1, params=dict(p=10))
                e = dag.df([[1, 2, 1, 10]], "a:int,ct1:int,ct2:int,p:int")
                e.assert_eq(c)

                a.zip(b).transform(mock_co_tf1, params=dict(p=10)).assert_eq(e)

                c = dag.transform(
                    a.zip(b), using=mock_co_tf1, params=dict(p=10, col="x")
                )
                e = dag.df([[1, 2, 1, 10]], "a:int,ct1:int,ct2:int,x:int")
                e.assert_eq(c)

                # interfaceless
                c = dag.transform(
                    a.zip(b),
                    using=mock_co_tf2,
                    schema="a:int,ct1:int,ct2:int,p:int",
                    params=dict(p=10),
                )
                e = dag.df([[1, 2, 1, 10]], "a:int,ct1:int,ct2:int,p:int")
                e.assert_eq(c)
                # single df
                c = dag.transform(a.zip(), using=mock_co_tf3)
                e = dag.df([[1, 3, 1]], "a:int,ct1:int,p:int")
                e.assert_eq(c)
                c = dag.transform(a.zip(partition=dict(by=["a"])), using=mock_co_tf3)
                e = dag.df([[1, 2, 1], [2, 1, 1]], "a:int,ct1:int,p:int")
                e.assert_eq(c)
                c = a.partition(by=["a"]).zip().transform(mock_co_tf3)
                e = dag.df([[1, 2, 1], [2, 1, 1]], "a:int,ct1:int,p:int")
                e.assert_eq(c)
                # ignore errors
                c = (
                    a.partition(by=["a"])
                    .zip()
                    .transform(mock_co_tf4_ex, ignore_errors=[NotImplementedError])
                )
                e = dag.df([[1, 2, 1]], "a:int,ct1:int,p:int")
                e.assert_eq(c)

        def test_cotransform_with_key(self):
            with self.dag() as dag:
                a = dag.df([[1, 2], [1, 3], [2, 1]], "a:int,b:int", dict(x=1))
                b = dag.df([[1, 2], [3, 4]], "a:int,c:int", dict(x=1))
                dag.zip(dict(x=a, y=b)).show()
                c = dag.zip(dict(x=a, y=b)).transform(
                    MockCoTransform1, params=dict(named=True)
                )
                e = dag.df([[1, 2, 1, 1]], "a:int,ct1:int,ct2:int,p:int")
                e.assert_eq(c)

                c = dag.zip(dict(df1=a, df2=b)).transform(
                    mock_co_tf1, params=dict(p=10)
                )
                e = dag.df([[1, 2, 1, 10]], "a:int,ct1:int,ct2:int,p:int")
                e.assert_eq(c)

                c = dag.zip(dict(df2=a, df1=b)).transform(
                    mock_co_tf1, params=dict(p=10)
                )
                e = dag.df([[1, 1, 2, 10]], "a:int,ct1:int,ct2:int,p:int")
                e.assert_eq(c)

                c = dag.transform(
                    dag.zip(dict(x=a, y=b)),
                    using=mock_co_tf2,
                    schema="a:int,ct1:int,ct2:int,p:int",
                    params=dict(p=10),
                )
                e = dag.df([[1, 2, 1, 10]], "a:int,ct1:int,ct2:int,p:int")
                e.assert_eq(c)

                c = dag.zip(dict(df1=a)).transform(mock_co_tf3)
                e = dag.df([[1, 3, 1]], "a:int,ct1:int,p:int")
                e.assert_eq(c)

        def test_out_transform(self):  # noqa: C901
            tmpdir = str(self.tmpdir)

            def incr():
                fs = FileSystem().makedirs(tmpdir, recreate=True)
                if fs.exists("a.txt"):
                    n = int(fs.readtext("a.txt"))
                else:
                    n = 0
                fs.writetext("a.txt", str(n + 1))
                return n

            def t1(df: Iterable[Dict[str, Any]]) -> Iterable[Dict[str, Any]]:
                for row in df:
                    incr()
                    yield row

            # partitionby_has: b
            def t2(df: pd.DataFrame) -> None:
                incr()

            # schema: *
            # partitionby_has: b
            def t3(df: pd.DataFrame) -> pd.DataFrame:
                incr()
                return df

            @transformer("*", partitionby_has=["b"])
            def t4(df: Iterable[Dict[str, Any]]) -> Iterable[Dict[str, Any]]:
                for row in df:
                    incr()
                    yield row

            @output_transformer(partitionby_has=["b"])
            def t5(df: Iterable[Dict[str, Any]]) -> Iterable[Dict[str, Any]]:
                for row in df:
                    incr()
                    yield row

            class T6(Transformer):
                def get_output_schema(self, df):
                    return df.schema

                def transform(self, df):
                    incr()
                    return df

            class T7(OutputTransformer):
                @property
                def validation_rules(self):
                    return {"partitionby_has": "b"}

                def process(self, df):
                    incr()

            def t8(df: pd.DataFrame) -> None:
                incr()
                raise NotImplementedError

            def t9(df: pd.DataFrame) -> Iterable[pd.DataFrame]:
                incr()
                yield df

            with self.dag() as dag:
                a = dag.df([[1, 2], [3, 4]], "a:double,b:int")
                a.out_transform(t1)  # +2
                a.partition(by=["b"]).out_transform(t2)  # +1 or +2
                a.partition(by=["b"]).out_transform(t3)  # +1 or +2
                a.partition(by=["b"]).out_transform(t4)  # +2
                a.partition(by=["b"]).out_transform(t5)  # +2
                a.out_transform(T6)  # +1
                a.partition(by=["b"]).out_transform(T7)  # +1
                a.out_transform(t8, ignore_errors=[NotImplementedError])  # +1
                a.out_transform(t9)  # +1
                raises(FugueWorkflowCompileValidationError, lambda: a.out_transform(t2))
                raises(FugueWorkflowCompileValidationError, lambda: a.out_transform(t3))
                raises(FugueWorkflowCompileValidationError, lambda: a.out_transform(t4))
                raises(FugueWorkflowCompileValidationError, lambda: a.out_transform(t5))
                raises(FugueWorkflowCompileValidationError, lambda: a.out_transform(T7))

            assert 12 <= incr()

        def test_output_cotransform(self):  # noqa: C901
            tmpdir = str(self.tmpdir)

            def incr():
                fs = FileSystem().makedirs(tmpdir, recreate=True)
                if fs.exists("a.txt"):
                    n = int(fs.readtext("a.txt"))
                else:
                    n = 0
                fs.writetext("a.txt", str(n + 1))
                return n

            def t1(
                df: Iterable[Dict[str, Any]], df2: pd.DataFrame
            ) -> Iterable[Dict[str, Any]]:
                for row in df:
                    incr()
                    yield row

            def t2(dfs: DataFrames) -> None:
                incr()

            def t3(df: pd.DataFrame, df2: pd.DataFrame) -> pd.DataFrame:
                incr()
                return df

            @cotransformer("a:double,b:int")
            def t4(
                df: Iterable[Dict[str, Any]], df2: pd.DataFrame
            ) -> Iterable[Dict[str, Any]]:
                for row in df:
                    incr()
                    yield row

            @output_cotransformer()
            def t5(
                df: Iterable[Dict[str, Any]], df2: pd.DataFrame
            ) -> Iterable[Dict[str, Any]]:
                for row in df:
                    incr()
                    yield row

            class T6(CoTransformer):
                def get_output_schema(self, dfs):
                    return dfs[0].schema

                def transform(self, dfs):
                    incr()
                    return dfs[0]

            class T7(OutputCoTransformer):
                def process(self, dfs):
                    incr()

            def t8(df: pd.DataFrame, df2: pd.DataFrame) -> pd.DataFrame:
                incr()
                raise NotImplementedError

            def t9(df1: pd.DataFrame, df2: pd.DataFrame) -> Iterable[pd.DataFrame]:
                incr()
                for df in [df1, df2]:
                    yield df

            with self.dag() as dag:
                a0 = dag.df([[1, 2], [3, 4]], "a:double,b:int")
                a1 = dag.df([[1, 2], [3, 4]], "aa:double,b:int")
                a = dag.zip(a0, a1)
                a.out_transform(t1)  # +2
                a.out_transform(t2)  # +1 or +2
                a.out_transform(t3)  # +1 or +2
                a.out_transform(t4)  # +2
                b = dag.zip(dict(df=a0, df2=a1))
                b.out_transform(t5)  # +2
                a.out_transform(T6)  # +1
                a.out_transform(T7)  # +1
                a.out_transform(t8, ignore_errors=[NotImplementedError])  # +1
                a.out_transform(t9)  # +1

            assert 12 <= incr()

        def test_join(self):
            with self.dag() as dag:
                a = dag.df([[1, 10], [2, 20], [3, 30]], "a:int,b:int")
                dag.join(a, how="inner").assert_eq(a)

                b = dag.df([[2, 200], [3, 300]], "a:int,c:int")
                c = dag.df([[2, 2000]], "a:int,d:int")
                d = a.join(b, c, how="inner")  # infer join on
                dag.df([[2, 20, 200, 2000]], "a:int,b:int,c:int,d:int").assert_eq(d)
                d = a.inner_join(b, c)
                dag.df([[2, 20, 200, 2000]], "a:int,b:int,c:int,d:int").assert_eq(d)
                d = a.semi_join(b, c)
                dag.df([[2, 20]], "a:int,b:int").assert_eq(d)
                d = a.left_semi_join(b, c)
                dag.df([[2, 20]], "a:int,b:int").assert_eq(d)
                d = a.anti_join(b, c)
                dag.df([[1, 10]], "a:int,b:int").assert_eq(d)
                d = a.left_anti_join(b, c)
                dag.df([[1, 10]], "a:int,b:int").assert_eq(d)

                # TODO: change these to str type to only test outer features?
                a = dag.df([[1, 10], [2, 20], [3, 30]], "a:int,b:int")
                b = dag.df([[2, 200], [3, 300]], "a:int,c:int")
                c = dag.df([[2, 2000], [4, 4000]], "a:int,d:int")
                d = a.left_outer_join(b, c)
                dag.df(
                    [[1, 10, None, None], [2, 20, 200, 2000], [3, 30, 300, None]],
                    "a:int,b:int,c:int,d:int",
                ).assert_eq(d)
                d = a.right_outer_join(b, c)
                dag.df(
                    [[2, 20, 200, 2000], [4, None, None, 4000]],
                    "a:int,b:int,c:int,d:int",
                ).assert_eq(d)
                d = a.full_outer_join(b, c)
                dag.df(
                    [
                        [1, 10, None, None],
                        [2, 20, 200, 2000],
                        [3, 30, 300, None],
                        [4, None, None, 4000],
                    ],
                    "a:int,b:int,c:int,d:int",
                ).assert_eq(d)

                a = dag.df([[1, 10], [2, 20]], "a:int,b:int")
                b = dag.df([[2], [3]], "c:int")
                c = dag.df([[4]], "d:int")
                d = a.cross_join(b, c)
                dag.df(
                    [[1, 10, 2, 4], [1, 10, 3, 4], [2, 20, 2, 4], [2, 20, 3, 4]],
                    "a:int,b:int,c:int,d:int",
                ).assert_eq(d)

        def test_select(self):
            class MockEngine(SqliteEngine):
                def __init__(self, execution_engine, p: int = 0):
                    super().__init__(execution_engine)
                    self.p = p

                def select(self, dfs, statement):
                    assert 2 == self.p  # assert set p value is working
                    return super().select(dfs, statement)

            with self.dag() as dag:
                a = dag.df([[1, 10], [2, 20], [3, 30]], "x:long,y:long")
                b = dag.df([[2, 20, 40], [3, 30, 90]], "x:long,y:long,z:long")
                dag.select("* FROM", a).assert_eq(a)
                dag.select("SELECT *,x*y AS z FROM", a, "WHERE x>=2").assert_eq(b)

                c = dag.df([[2, 20, 40], [3, 30, 90]], "x:long,y:long,zb:long")
                dag.select(
                    "  SELECT t1.*,z AS zb FROM ",
                    a,
                    "AS t1 INNER JOIN",
                    b,
                    "AS t2 ON t1.x=t2.x  ",
                ).assert_eq(c)

                # no select
                dag.select(
                    "t1.*,z AS zb FROM ", a, "AS t1 INNER JOIN", b, "AS t2 ON t1.x=t2.x"
                ).assert_eq(c)

                # specify sql engine
                dag.select(
                    "SELECT t1.*,z AS zb FROM ",
                    a,
                    "AS t1 INNER JOIN",
                    b,
                    "AS t2 ON t1.x=t2.x",
                    sql_engine=SqliteEngine,
                ).assert_eq(c)

                # specify sql engine and params
                dag.select(
                    "SELECT t1.*,z AS zb FROM ",
                    a,
                    "AS t1 INNER JOIN",
                    b,
                    "AS t2 ON t1.x=t2.x",
                    sql_engine=MockEngine,
                    sql_engine_params={"p": 2},
                ).assert_eq(c)

                # no input
                dag.select("9223372036854775807 AS a").assert_eq(
                    dag.df([[9223372036854775807]], "a:long")
                )

                # make sure transform -> select works
                b = a.transform(mock_tf1)
                a = a.transform(mock_tf1)
                aa = dag.select("* FROM", a)
                dag.select("* FROM", b).assert_eq(aa)

        def test_union(self):
            with self.dag() as dag:
                a = dag.df([[1, 10], [2, None], [2, None]], "x:long,y:double")
                b = dag.df([[2, None], [2, 20]], "x:long,y:double")
                c = dag.df([[1, 10], [2, 20]], "x:long,y:double")
                a.union().assert_eq(a)
                a.union(b, c).assert_eq(
                    ArrayDataFrame(
                        [
                            [1, 10],
                            [2, None],
                            [2, 20],
                        ],
                        "x:long,y:double",
                    )
                )
                a.union(b, c, distinct=False).assert_eq(
                    ArrayDataFrame(
                        [
                            [1, 10],
                            [2, None],
                            [2, None],
                            [2, None],
                            [2, 20],
                            [1, 10],
                            [2, 20],
                        ],
                        "x:long,y:double",
                    )
                )

        def test_intersect(self):
            with self.dag() as dag:
                a = dag.df([[1, 10], [2, None], [2, None]], "x:long,y:double")
                b = dag.df([[2, None], [2, 20]], "x:long,y:double")
                c = dag.df([[1, 10], [2, 20]], "x:long,y:double")
                # d = dag.df([[1, 10], [2, 20], [2, None]], "x:long,y:double")
                a.intersect(b).assert_eq(
                    ArrayDataFrame(
                        [[2, None]],
                        "x:long,y:double",
                    )
                )
                a.intersect(b, c).assert_eq(
                    ArrayDataFrame(
                        [],
                        "x:long,y:double",
                    )
                )
                # TODO: INTERSECT ALL is not implemented (QPD issue)
                # a.intersect(b, distinct=False).assert_eq(
                #     ArrayDataFrame(
                #         [[2, None], [2, None]],
                #         "x:long,y:double",
                #     )
                # )
                # a.intersect(b, d, distinct=False).assert_eq(
                #     ArrayDataFrame(
                #         [[2, None], [2, None]],
                #         "x:long,y:double",
                #     )
                # )

        def test_subtract(self):
            with self.dag() as dag:
                a = dag.df([[1, 10], [2, None], [2, None]], "x:long,y:double")
                b = dag.df([[2, None], [2, 20]], "x:long,y:double")
                c = dag.df([[1, 10], [2, 20]], "x:long,y:double")
                a.subtract(b).assert_eq(
                    ArrayDataFrame(
                        [[1, 10]],
                        "x:long,y:double",
                    )
                )
                a.subtract(c).assert_eq(
                    ArrayDataFrame(
                        [[2, None]],
                        "x:long,y:double",
                    )
                )
                # # TODO: EXCEPT ALL is not implemented (QPD issue)
                # a.subtract(c, distinct=False).assert_eq(
                #     ArrayDataFrame(
                #         [[2, None], [2, None]],
                #         "x:long,y:double",
                #     )
                # )
                a.subtract(b, c).assert_eq(
                    ArrayDataFrame(
                        [],
                        "x:long,y:double",
                    )
                )

        def test_distinct(self):
            with self.dag() as dag:
                a = dag.df([[1, 10], [2, None], [2, None]], "x:long,y:double")
                a.distinct().assert_eq(
                    ArrayDataFrame(
                        [[1, 10], [2, None]],
                        "x:long,y:double",
                    )
                )

        def test_dropna(self):
            with self.dag() as dag:
                a = dag.df(
                    [[1, 10, 10], [None, 2, None], [2, None, 4]],
                    "x:double,y:double,z:double",
                )
                a.dropna().assert_eq(
                    ArrayDataFrame(
                        [[1, 10, 10]],
                        "x:double,y:double,z:double",
                    )
                )
                a.dropna(how="all").assert_eq(a)
                a.dropna(thresh=2).assert_eq(
                    ArrayDataFrame(
                        [[1, 10, 10], [2, None, 4]],
                        "x:double,y:double,z:double",
                    )
                )
                a.dropna(how="any", subset=["x", "z"]).assert_eq(
                    ArrayDataFrame(
                        [[1, 10, 10], [2, None, 4]],
                        "x:double,y:double,z:double",
                    )
                )
                a.dropna(thresh=1, subset=["y", "z"]).assert_eq(a)

        def test_fillna(self):
            with self.dag() as dag:
                a = dag.df(
                    [[1, 10, 10], [None, 2, None], [2, None, 4]],
                    "x:double,y:double,z:double",
                )
                a.fillna(-99).assert_eq(
                    ArrayDataFrame(
                        [[1, 10, 10], [-99, 2, -99], [2, -99, 4]],
                        "x:double,y:double,z:double",
                    )
                )
                a.fillna(-99, subset=["y"]).assert_eq(
                    ArrayDataFrame(
                        [[1, 10, 10], [None, 2, None], [2, -99, 4]],
                        "x:double,y:double,z:double",
                    )
                )
                # Subset is ignored if value is a dict
                a.fillna({"y": 0, "z": -99}, subset=["y"]).assert_eq(
                    ArrayDataFrame(
                        [[1, 10, 10], [None, 2, -99], [2, 0, 4]],
                        "x:double,y:double,z:double",
                    )
                )
            with raises(FugueWorkflowError):
                with self.dag() as dag:
                    dag.df([[None, 1]], "a:int,b:int").fillna({"a": None, "b": 1})

            with raises(FugueWorkflowError):
                with self.dag() as dag:
                    dag.df([[None, 1]], "a:int,b:int").fillna(None)

        def test_sample(self):
            with self.dag() as dag:
                a = dag.df(
                    [[1, 10, 10], [None, 2, None], [2, None, 4]],
                    "x:double,y:double,z:double",
                )
                a.sample(frac=0.5, replace=False, seed=0).show()

            with self.dag() as dag:
                with raises(ValueError):
                    dag.df([[None, 1]], "a:int,b:int").sample(n=1, frac=0.2)

        def test_col_ops(self):
            with self.dag() as dag:
                a = dag.df([[1, 10], [2, 20]], "x:long,y:long")
                aa = dag.df([[1, 10], [2, 20]], "xx:long,y:long")
                a.rename({"x": "xx"}).assert_eq(aa)
                a[["x"]].assert_eq(ArrayDataFrame([[1], [2]], "x:long"))

                a.drop(["y", "yy"], if_exists=True).assert_eq(
                    ArrayDataFrame([[1], [2]], "x:long")
                )

                a[["x"]].rename(x="xx").assert_eq(ArrayDataFrame([[1], [2]], "xx:long"))
                a.alter_columns("x:str").assert_eq(
                    ArrayDataFrame([["1", 10], ["2", 20]], "x:str,y:long")
                )

        def test_datetime_in_workflow(self):
            # schema: a:date,b:datetime
            def t1(df: pd.DataFrame) -> pd.DataFrame:
                df["b"] = "2020-01-02"
                df["b"] = pd.to_datetime(df["b"])
                return df

            class T2(Transformer):
                def get_output_schema(self, df):
                    return df.schema

                def transform(self, df):
                    # test for issue https://github.com/fugue-project/fugue/issues/92
                    return PandasDataFrame(df.as_pandas())

            with self.dag() as dag:
                a = dag.df([["2020-01-01"]], "a:date").transform(t1)
                b = dag.df(
                    [[datetime.date(2020, 1, 1), datetime.datetime(2020, 1, 2)]],
                    "a:date,b:datetime",
                )
                b.assert_eq(a)
                c = dag.df([["2020-01-01", "2020-01-01 00:00:00"]], "a:date,b:datetime")
                c.transform(T2).assert_eq(c)
                c.partition(by=["a"]).transform(T2).assert_eq(c)

        @pytest.fixture(autouse=True)
        def init_tmpdir(self, tmpdir):
            self.tmpdir = tmpdir

        def test_io(self):
            path = os.path.join(self.tmpdir, "a")
            path2 = os.path.join(self.tmpdir, "b.test.csv")
            with self.dag() as dag:
                b = dag.df([[6, 1], [2, 7]], "c:int,a:long")
                b.partition(num=3).save(path, fmt="parquet", single=True)
                b.save(path2, header=True)
            assert FileSystem().isfile(path)
            with self.dag() as dag:
                a = dag.load(path, fmt="parquet", columns=["a", "c"])
                a.assert_eq(dag.df([[1, 6], [7, 2]], "a:long,c:int"))
                a = dag.load(path2, header=True, columns="c:int,a:long")
                a.assert_eq(dag.df([[6, 1], [2, 7]], "c:int,a:long"))

        def test_save_and_use(self):
            path = os.path.join(self.tmpdir, "a")
            with self.dag() as dag:
                b = dag.df([[6, 1], [2, 7]], "c:int,a:long")
                c = b.save_and_use(path, fmt="parquet")
                b.assert_eq(c)

            with self.dag() as dag:
                b = dag.df([[6, 1], [2, 7]], "c:int,a:long")
                d = dag.load(path, fmt="parquet")
                b.assert_eq(d)

        def test_transformer_validation(self):
            # partitionby_has: b
            # input_has: a
            # schema: *
            def t1(df: pd.DataFrame) -> pd.DataFrame:
                return df

            @transformer("*", partitionby_has=["b"], input_has=["a"])
            def t2(df: pd.DataFrame) -> pd.DataFrame:
                return df

            class T3(Transformer):
                @property
                def validation_rules(self):
                    return dict(partitionby_has=["b"], input_has=["a"])

                def get_output_schema(self, df: DataFrame) -> Any:
                    return df.schema

                def transform(self, df: DataFrame) -> LocalDataFrame:
                    return df.as_local()

            for t in [t1, t2, T3]:
                # compile time
                with raises(FugueWorkflowCompileValidationError):
                    self.dag().df([[0, 1]], "a:int,b:int").transform(t)

                # runtime
                with raises(FugueWorkflowRuntimeValidationError):
                    with self.dag() as dag:
                        dag.df([[0, 1]], "c:int,b:int").partition(by=["b"]).transform(t)

                with self.dag() as dag:
                    dag.df([[0, 1]], "a:int,b:int").partition(by=["b"]).transform(
                        t
                    ).assert_eq(dag.df([[0, 1]], "a:int,b:int"))

        def test_processor_validation(self):
            # input_has: a
            def p1(dfs: DataFrames) -> DataFrame:
                return dfs[0]

            @processor(input_has=["a"])
            def p2(dfs: DataFrames) -> DataFrame:
                return dfs[0]

            class P3(Processor):
                @property
                def validation_rules(self):
                    return dict(input_has=["a"])

                def process(self, dfs: DataFrames) -> DataFrame:
                    return dfs[0]

            for p in [p1, p2, P3]:
                # run time
                with raises(FugueWorkflowRuntimeValidationError):
                    with self.dag() as dag:
                        df1 = dag.df([[0, 1]], "a:int,b:int")
                        df2 = dag.df([[0, 1]], "c:int,d:int")
                        dag.process([df1, df2], using=p)

                with self.dag() as dag:
                    df1 = dag.df([[0, 1]], "a:int,b:int")
                    df2 = dag.df([[0, 1]], "a:int,b:int")
                    dag.process([df1, df2], using=p).assert_eq(df1)

            # input_has: a
            # partitionby_has: b
            def p4(dfs: DataFrames) -> DataFrame:
                return dfs[0]

            # compile time
            with raises(FugueWorkflowCompileValidationError):
                dag = self.dag()
                df = dag.df([[0, 1]], "a:int,b:int")
                df.process(p4)

            with self.dag() as dag:
                dag.df([[0, 1]], "a:int,b:int").partition(by=["b"]).process(p4)

        def test_outputter_validation(self):
            # input_has: a
            def o1(dfs: DataFrames) -> None:
                pass

            @outputter(input_has=["a"])
            def o2(dfs: DataFrames) -> None:
                pass

            class O3(Outputter):
                @property
                def validation_rules(self):
                    return dict(input_has=["a"])

                def process(self, dfs: DataFrames) -> None:
                    pass

            for o in [o1, o2, O3]:
                # run time
                with raises(FugueWorkflowRuntimeValidationError):
                    with self.dag() as dag:
                        df1 = dag.df([[0, 1]], "a:int,b:int")
                        df2 = dag.df([[0, 1]], "c:int,d:int")
                        dag.output([df1, df2], using=o)

                with self.dag() as dag:
                    df1 = dag.df([[0, 1]], "a:int,b:int")
                    df2 = dag.df([[0, 1]], "a:int,b:int")
                    dag.output([df1, df2], using=o)

            # input_has: a
            # partitionby_has: b
            def o4(dfs: DataFrames) -> None:
                pass

            # compile time
            with raises(FugueWorkflowCompileValidationError):
                dag = self.dag()
                df = dag.df([[0, 1]], "a:int,b:int")
                df.output(o4)

            with self.dag() as dag:
                dag.df([[0, 1]], "a:int,b:int").partition(by=["b"]).output(o4)


def mock_creator(p: int) -> DataFrame:
    return ArrayDataFrame([[p]], "a:int")


def mock_processor(df1: List[List[Any]], df2: List[List[Any]]) -> DataFrame:
    return ArrayDataFrame([[len(df1) + len(df2)]], "a:int")


def mock_processor2(e: ExecutionEngine, dfs: DataFrames) -> DataFrame:
    assert "test" in e.conf
    return ArrayDataFrame([[sum(s.count() for s in dfs.values())]], "a:int")


class MockProcessor3(Processor):
    def process(self, dfs):
        assert "test" in self.workflow_conf
        return ArrayDataFrame([[sum(s.count() for s in dfs.values())]], "a:int")


def mock_outputter(df1: List[List[Any]], df2: List[List[Any]]) -> None:
    assert len(df1) == len(df2)


def mock_outputter2(df: List[List[Any]]) -> None:
    print(df)


class MockOutputter3(Outputter):
    def process(self, dfs):
        assert "3" == self.partition_spec.num_partitions


class MockOutputter4(Outputter):
    def process(self, dfs):
        for k, v in dfs.items():
            print(k)
            v.show()


class MockTransform1(Transformer):
    def get_output_schema(self, df: DataFrame) -> Any:
        assert "test" in self.workflow_conf
        assert "x" in df.metadata
        return [df.schema, "ct:int,p:int"]

    def on_init(self, df: DataFrame) -> None:
        assert "test" in self.workflow_conf
        assert "x" in df.metadata
        self.pn = self.cursor.physical_partition_no
        self.ks = self.key_schema
        if "on_init_called" not in self.__dict__:
            self.on_init_called = 1
        else:  # pragma: no cover
            self.on_init_called += 1

    def transform(self, df: LocalDataFrame) -> LocalDataFrame:
        assert 1 == self.on_init_called
        assert "test" in self.workflow_conf
        assert "x" in df.metadata
        pdf = df.as_pandas()
        pdf["p"] = self.params.get("p", 1)
        pdf["ct"] = pdf.shape[0]
        return PandasDataFrame(pdf, self.output_schema)


@transformer(lambda df, **kwargs: df.schema + (kwargs.get("col", "p") + ":int"))
def mock_tf0(df: pd.DataFrame, p=1, col="p") -> pd.DataFrame:
    df[col] = p
    return df


# schema: *,ct:int,p:int
def mock_tf1(df: pd.DataFrame, p=1) -> pd.DataFrame:
    df["ct"] = df.shape[0]
    df["p"] = p
    return df


def mock_tf2_except(df: Iterable[Dict[str, Any]], p=1) -> Iterable[Dict[str, Any]]:
    n = 0
    for row in df:
        yield row
        n += 1
        if n > 1:
            raise NotImplementedError


# schema: *
def mock_tf3(df: Iterable[Dict[str, Any]]) -> Iterable[Dict[str, Any]]:
    # binary data test
    for row in df:
        obj = pickle.loads(row["b"])
        obj[0] += 1
        obj[1] += "x"
        row["b"] = pickle.dumps(obj)
        yield row


class MockCoTransform1(CoTransformer):
    def get_output_schema(self, dfs: DataFrames) -> Any:
        assert "test" in self.workflow_conf
        assert 2 == len(dfs)
        if self.params.get("named", False):
            assert dfs.has_key
        else:
            assert not dfs.has_key
        return [self.key_schema, "ct1:int,ct2:int,p:int"]

    def on_init(self, dfs: DataFrames) -> None:
        assert "test" in self.workflow_conf
        assert 2 == len(dfs)
        if self.params.get("named", False):
            assert dfs.has_key
        else:
            assert not dfs.has_key
        self.pn = self.cursor.physical_partition_no
        self.ks = self.key_schema
        if "on_init_called" not in self.__dict__:
            self.on_init_called = 1
        else:  # pragma: no cover
            self.on_init_called += 1

    def transform(self, dfs: DataFrames) -> LocalDataFrame:
        assert 1 == self.on_init_called
        assert "test" in self.workflow_conf
        assert 2 == len(dfs)
        if self.params.get("named", False):
            assert dfs.has_key
        else:
            assert not dfs.has_key
        row = self.cursor.key_value_array + [
            dfs[0].count(),
            dfs[1].count(),
            self.params.get("p", 1),
        ]
        return ArrayDataFrame([row], self.output_schema)


@cotransformer(
    lambda dfs, **kwargs: "a:int,ct1:int,ct2:int," + kwargs.get("col", "p") + ":int"
)
def mock_co_tf1(
    df1: List[Dict[str, Any]], df2: List[List[Any]], p=1, col="p"
) -> List[List[Any]]:
    return [[df1[0]["a"], len(df1), len(df2), p]]


def mock_co_tf2(dfs: DataFrames, p=1) -> List[List[Any]]:
    return [[dfs[0].peek_dict()["a"], dfs[0].count(), dfs[1].count(), p]]


@cotransformer(Schema("a:int,ct1:int,p:int"))
def mock_co_tf3(df1: List[Dict[str, Any]], p=1) -> List[List[Any]]:
    return [[df1[0]["a"], len(df1), p]]


@cotransformer("a:int,ct1:int,p:int")
def mock_co_tf4_ex(df1: List[Dict[str, Any]], p=1) -> List[List[Any]]:
    k = df1[0]["a"]
    if k == 2:
        raise NotImplementedError
    return [[df1[0]["a"], len(df1), p]]
